"""Module to read and return anatomical information, such as:
    - surfaces, with class Surf
    - brains, with class BrainSurf (both hemispheres)
"""
from collections import Counter
from logging import getLogger
from os import environ
from pathlib import Path
from re import compile
from struct import unpack

from numpy import (array, empty, vstack, around, dot, append, reshape,
                   meshgrid, asarray)

lg = getLogger(__name__)

try:
    from nibabel.freesurfer import load, read_annot
    from nibabel import load as nload
except ImportError:
    pass

FS_AFFINE = array([[-1, 0, 0, 128],
                   [0, 0, -1, 128],
                   [0, 1, 0, 128],
                   [0, 0, 0, 1]])
HEMISPHERES = 'lh', 'rh'


def _read_geometry(surf_file):
    """Read a triangular format Freesurfer surface mesh.

    Parameters
    ----------
    surf_file : str
        path to surface file

    Returns
    -------
    coords : numpy.ndarray
        nvtx x 3 array of vertex (x, y, z) coordinates
    faces : numpy.ndarray
        nfaces x 3 array of defining mesh triangles

    Notes
    -----
    This function comes from nibabel, but it doesn't use numpy because numpy
    doesn't return the correct values in Python 3.
    """
    with open(surf_file, 'rb') as f:
        filebytes = f.read()

    assert filebytes[:3] == b'\xff\xff\xfe'
    i0 = filebytes.index(b'\x0A\x0A') + 2
    i1 = i0 + 4
    vnum = unpack('>i', filebytes[i0:i1])[0]
    i0 = i1
    i1 += 4
    fnum = unpack('>i', filebytes[i0:i1])[0]
    i0 = i1
    i1 += 4 * vnum * 3
    verts = unpack('>' + 'f' * vnum * 3, filebytes[i0:i1])
    i0 = i1
    i1 += 4 * fnum * 3
    faces = unpack('>' + 'i' * fnum * 3, filebytes[i0:i1])

    verts = asarray(verts).reshape(vnum, 3)
    faces = asarray(faces).reshape(fnum, 3)
    return verts, faces


def _find_neighboring_regions(pos, mri_dat, region, approx, exclude_regions):

    spot_size = approx * 2 + 1

    x, y, z = meshgrid(range(spot_size), range(spot_size), range(spot_size))
    neighb = vstack((reshape(x, (1, spot_size ** 3)),
                     reshape(y, (1, spot_size ** 3)),
                     reshape(z, (1, spot_size ** 3)))).T - approx
    regions = []
    for p in range(neighb.shape[0]):
        d_type = mri_dat[pos[0] + neighb[p, 0], pos[1] + neighb[p, 1],
                         pos[2] + neighb[p, 2]]
        label_index = region['index'].index(d_type)
        regions.append(region['label'][label_index])

    if exclude_regions:
        excluded = compile('|'.join(exclude_regions))
        regions = [x for x in regions if not excluded.search(x)]

    return regions


def import_freesurfer_LUT(fs_lut=None):
    """Import Look-up Table with colors and labels for anatomical regions.

    It's necessary that Freesurfer is installed and that the environmental
    variable 'FREESURFER_HOME' is present.

    Parameters
    ----------
    fs_lut : str or Path
        path to file called FreeSurferColorLUT.txt

    Returns
    -------
    idx : list of int
        indices of regions
    label : list of str
        names of the brain regions
    rgba : numpy.ndarray
        one row is a brain region and the columns are the RGB + alpha colors
    """
    if fs_lut is not None:
        lg.info('Reading user-specified lookuptable {}'.format(fs_lut))
    else:
        try:
            fs_home = environ['FREESURFER_HOME']
        except KeyError:
            raise OSError('Freesurfer is not installed or FREESURFER_HOME is '
                          'not defined as environmental variable')
        else:
            fs_lut = Path(fs_home) / 'FreeSurferColorLUT.txt'
            lg.info('Reading lookuptable in FREESURFER_HOME {}'.format(fs_lut))

    idx = []
    label = []
    rgba = empty((0, 4))
    with open(fs_lut, 'r') as f:
        for l in f:
            if len(l) <= 1 or l[0] == '#' or l[0] == '\r':
                continue
            (t0, t1, t2, t3, t4, t5) = [t(s) for t, s in
                                        zip((int, str, int, int, int, int),
                                        l.split())]
            idx.append(t0)
            label.append(t1)
            rgba = vstack((rgba, array([t2, t3, t4, t5])))

    return idx, label, rgba


class Surf:
    """Provide class Surf, with the positions and connections of vertices.

    Parameters
    ----------
    surf_file : str or Path
        freesurfer file containing the surface

    Attributes
    ----------
    surf_file : path to file
        freesurfer file containing the surface
    vert : numpy.ndarray
        vertices of the mesh
    tri : numpy.ndarray
        triangulation of the mesh
    """
    def __init__(self, surf_file):
        self.surf_file = surf_file
        surf_vert, surf_tri = _read_geometry(self.surf_file)
        self.vert = surf_vert
        self.tri = surf_tri
        self.n_vert = surf_vert.shape[0]


class Brain:
    """Class that contains the left and right hemispheres.

    Parameters
    ----------
    freesurfer_dir : str
        subject-specific directory created by freesurfer
    surf_type : str
        'pial', 'smoothwm', 'inflated', 'white', or 'sphere'
    """
    def __init__(self, freesurfer_dir, surf_type='pial'):

        freesurfer_dir = Path(freesurfer_dir)

        for hemi in HEMISPHERES:
            surf_file = freesurfer_dir / 'surf' / (hemi + '.' + surf_type)
            setattr(self, hemi, Surf(surf_file))


class Freesurfer:
    """Provide class Freesurfer, with the information from freesurfer.

    Parameters
    ----------
    freesurfer_dir : str
        subject-specific directory created by freesurfer
    fs_lut : str
        path to file called FreeSurferColorLUT.txt

    Notes
    -----
    It's necessary that Freesurfer is installed and that the environmental
    variable 'FREESURFER_HOME' is present.
    """
    def __init__(self, freesurfer_dir, fs_lut=None):
        freesurfer_dir = Path(freesurfer_dir)
        if not freesurfer_dir.exists():
            raise OSError(str(freesurfer_dir) + ' does not exist')

        self.dir = freesurfer_dir
        try:
            lut = import_freesurfer_LUT(fs_lut)
            self.lookuptable = {'index': lut[0], 'label': lut[1],
                                'RGBA': lut[2]}
        except OSError as err:
            self.lookuptable = None
            lg.warning('Could not find lookup table (see below for explanation)'
                       '. Some functions that rely on it might complain or '
                       'crash.')
            lg.warning(err)

    @property
    def surface_ras_shift(self):
        """Freesurfer uses two coordinate systems: one for volumes ("RAS") and
        one for surfaces ("tkReg", "tkRAS", and "Surface RAS").
        To get from surface to volume coordinates, add this numbers.
        To get from volume to surface coordinates, substract this numbers.
        """
        T1_path = self.dir / 'mri' / 'T1.mgz'
        assert T1_path.exists()

        try:
            T1 = nload(str(T1_path))
        except NameError:
            raise ImportError('nibabel needs to be installed for this function')

        return T1.header['Pxyz_c']

    def find_brain_region(self, abs_pos, parc_type='aparc', max_approx=None,
                          exclude_regions=None):
        """Find the name of the brain region in which an electrode is located.

        Parameters
        ----------
        abs_pos : numpy.ndarray
            3x0 vector with the position of interest.
        parc_type : str
            'aparc', 'aparc.a2009s', 'BA', 'BA.thresh', or 'aparc.DKTatlas40'
            'aparc.DKTatlas40' is only for recent freesurfer versions
        max_approx : int, optional
            max approximation to define position of the electrode.
        exclude_regions : list of str or empty list
            do not report regions if they contain these substrings. None means
            that it does not exclude any region.

        Notes
        -----
        It determines the possible brain region in which one electrode is
        present, based on Freesurfer segmentation. You can imagine that
        sometimes one electrode is not perfectly located within one region,
        but it's a few mm away. The parameter "approx" specifies this tolerance
        where each value is one mm. It keeps on searching in larger and larger
        spots until it finds at least one region which is not white matter. If
        there are multiple regions, it returns the region with the most
        detection.
        Minimal value is 0, which means only if the electrode is in the
        precise location.

        If you want to exclude white matter regions with 'aparc', use
            exclude_regions = ('White', 'WM', 'Unknown')
        and with 'aparc.a2009s', use:
            exclude_regions = ('White-Matter')
        """
        # convert to freesurfer coordinates of the MRI
        pos = around(dot(FS_AFFINE, append(abs_pos, 1)))[:3].astype(int)
        lg.debug('Position in the MRI matrix: {}'.format(pos))

        mri_dat, _ = self.read_seg(parc_type)

        if max_approx is None:
            max_approx = 3

        for approx in range(max_approx + 1):
            lg.debug('Trying approx {} out of {}'.format(approx, max_approx))
            regions = _find_neighboring_regions(pos, mri_dat,
                                                self.lookuptable, approx,
                                                exclude_regions)
            if regions:
                break

        if regions:
            c_regions = Counter(regions)
            return c_regions.most_common(1)[0][0], approx
        else:
            return '--not found--', approx

    def read_label(self, hemi, parc_type='aparc'):
        """Read the labels (annotations) for each hemisphere.

        Parameters
        ----------
        hemi : str
            'lh' or 'rh'
        parc_type : str
            'aparc', 'aparc.a2009s', 'BA', 'BA.thresh', or 'aparc.DKTatlas40'
            'aparc.DKTatlas40' is only for recent freesurfer versions

        Returns
        -------
        numpy.ndarray
            value at each vertex, indicating the label
        numpy.ndarray
            RGB + alpha colors for each label
        list of str
            names of the labels
        """
        parc_file = self.dir / 'label' / (hemi + '.' + parc_type + '.annot')
        try:
            vert_val, region_color, region_name = read_annot(parc_file)
        except NameError:
            raise ImportError('nibabel needs to be installed for this function')
        region_name = [x.decode('utf-8') for x in region_name]
        return vert_val, region_color, region_name

    def read_seg(self, parc_type='aparc'):
        """Read the MRI segmentation.

        Parameters
        ----------
        parc_type : str
            'aparc' or 'aparc.a2009s'

        Returns
        -------
        numpy.ndarray
            3d matrix with values
        numpy.ndarray
            4x4 affine matrix
        """
        seg_file = self.dir / 'mri' / (parc_type + '+aseg.mgz')
        try:
            seg_mri = load(seg_file)
        except NameError:
            raise ImportError('nibabel needs to be installed for this function')
        seg_aff = seg_mri.affine
        seg_dat = seg_mri.get_data()
        return seg_dat, seg_aff

    def read_brain(self, surf_type='pial'):
        """Read the surface of both hemispheres.

        Parameters
        ----------
        surf_type : str
            'pial', 'smoothwm', 'inflated', 'white', or 'sphere'

        Returns
        -------
        instance of Brain
            the surfaces of both brain hemispheres
        """
        return Brain(self.dir, surf_type=surf_type)



from numpy.testing import assert_array_equal

from wonambi import Dataset
from wonambi.ioeeg import write_wonambi
from wonambi.utils import create_data

from .paths import wonambi_file

gen_data = create_data(n_trial=1)


def test_wonambi_write_read():
    write_wonambi(gen_data, wonambi_file, subj_id='test_subj')
    d = Dataset(wonambi_file)
    data = d.read_data()
    assert_array_equal(data(trial=0), gen_data(trial=0))


from pytest import raises
from wonambi.utils import create_data


def test_import():
    import wonambi
    print(wonambi)


def test_simulate_01():
    data = create_data()
    assert data.data.dtype == 'O'
    assert data.data.shape == (1,)  # one trial
    assert data.data[0].shape[0] == len(data.axis['chan'][0])
    assert data.data[0].shape[1] == len(data.axis['time'][0])


def test_simulate_02():
    with raises(ValueError):
        create_data(datatype='xxx')


def test_simulate_03():
    N_TRIAL = 10
    data = create_data(n_trial=N_TRIAL)
    assert data.data.shape[0] == N_TRIAL


def test_simulate_04():
    data = create_data(datatype='ChanFreq')
    assert data.data[0].shape[0] == len(data.axis['chan'][0])
    assert data.data[0].shape[1] == len(data.axis['freq'][0])


def test_simulate_05():
    data = create_data(datatype='ChanTimeFreq')
    assert data.data[0].shape[0] == len(data.axis['chan'][0])
    assert data.data[0].shape[1] == len(data.axis['time'][0])
    assert data.data[0].shape[2] == len(data.axis['freq'][0])


def test_simulate_06():
    TIME_LIMITS = (0, 10)
    data = create_data(time=TIME_LIMITS)
    assert data.axis['time'][0][0] == TIME_LIMITS[0]
    assert data.axis['time'][0][-1] < TIME_LIMITS[1]


def test_simulate_07():
    FREQ_LIMITS = (0, 10)
    data = create_data(datatype='ChanFreq', freq=FREQ_LIMITS)
    assert data.axis['freq'][0][0] == FREQ_LIMITS[0]
    assert data.axis['freq'][0][-1] < FREQ_LIMITS[1]



"""Package containing Qt widgets.

The program is organized in widgets:
  - info/Info : Information about the dataset.
  - overview/Overview : Show an overview of data.
  - traces/Traces : Show the traces of activity.
  - labels/Labels : Allow user to change labels.
  - channels/Channels : Allow user to choose channels, and filters.
  - notes/Notes : Show annotations, such as markers and sleep staging
  - spectrum/Spectrum : Show the Welch's periodogram of one channel.
  - video/Video : Show the video of the recordings.

A module should start with:

from logging import getLogger
lg = getLogger(__name__)

then import packages that are in the Python library, additional packages (in
this order: numpy, scipy, PyQt5), relative imports (leave an empty
line between each group of imports).

"""



"""Module to show video, if the file format supports it.

"""
from logging import getLogger
from platform import system

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import (QFrame,
                             QGroupBox,
                             QPushButton,
                             QVBoxLayout,
                             QWidget,
                             )

try:
    import vlc
except:
    vlc = False

from .settings import Config

lg = getLogger(__name__)


class ConfigVideo(Config):

    def __init__(self):
        super().__init__('video', None)

    def create_config(self):

        box0 = QGroupBox('Video')

        main_layout = QVBoxLayout()
        main_layout.addWidget(box0)

        self.setLayout(main_layout)


class Video(QWidget):
    """Widget containing the movie, if available.

    Attributes
    ----------
    parent : instance of QMainWindow
        the main window.
    beg_diff : float
        time in ms of the beginning of the first video
    end_diff : float
        time in ms of the end of the last video
    cnt_video : int
        index of the current mediasource
    n_video : int
        total number of videos to play
    video : instance of MediaObject
        the video to show.
    idx_button : instance of QPushButton
        button which starts and stops the video.
    """
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.config = ConfigVideo()

        self.cnt_video = 0
        self.n_video = 0
        self.mediaplayer = None

        self.idx_button = None

        if vlc:
            self.create_video()

    def create_video(self):
        """Create video widget."""

        self.instance = vlc.Instance()

        video_widget = QFrame()
        self.mediaplayer = self.instance.media_player_new()
        if system() == 'Linux':
            self.mediaplayer.set_xwindow(video_widget.winId())
        elif system() == 'Windows':
            self.mediaplayer.set_hwnd(video_widget.winId())
        elif system() == 'darwin':  # to test
            self.mediaplayer.set_nsobject(video_widget.winId())
        else:
            lg.warning('unsupported system for video widget')
            return

        self.medialistplayer = vlc.MediaListPlayer()
        self.medialistplayer.set_media_player(self.mediaplayer)
        event_manager = self.medialistplayer.event_manager()
        event_manager.event_attach(vlc.EventType.MediaListPlayerNextItemSet,
                                   self.next_video)

        self.idx_button = QPushButton()
        self.idx_button.setText('Start')
        self.idx_button.clicked.connect(self.start_stop_video)

        layout = QVBoxLayout()
        layout.addWidget(video_widget)
        layout.addWidget(self.idx_button)
        self.setLayout(layout)

    def stop_video(self, tick):
        """Stop video if tick is more than the end, only for last file.

        Parameters
        ----------
        tick : int
            time in ms from the beginning of the file

        useless?
        """
        if self.cnt_video == self.n_video:
            if tick >= self.end_diff:
                self.idx_button.setText('Start')
                self.video.stop()

    def check_if_finished(self):
        if self.cnt_video == self.n_video:
            if self.mediaplayer.get_time() > self.endsec * 1000:
                self.medialistplayer.stop()
                self.t.stop()

    def next_video(self, _):
        """Also runs when file is loaded, so index starts at 2."""
        self.cnt_video += 1
        lg.info('Update video to ' + str(self.cnt_video))

    def start_stop_video(self):
        """Start and stop the video, and change the button.
        """
        if self.parent.info.dataset is None:
            self.parent.statusBar().showMessage('No Dataset Loaded')
            return

        # & is added automatically by PyQt, it seems
        if 'Start' in self.idx_button.text().replace('&', ''):
            try:
                self.update_video()
            except IndexError as er:
                lg.debug(er)
                self.idx_button.setText('Not Available / Start')
                return
            except OSError as er:
                lg.debug(er)
                self.idx_button.setText('NO VIDEO for this dataset')
                return

            self.idx_button.setText('Stop')

        elif 'Stop' in self.idx_button.text():
            self.idx_button.setText('Start')
            self.medialistplayer.stop()

    def update_video(self):
        """Read list of files, convert to video time, and add video to queue.
        """
        window_start = self.parent.value('window_start')
        window_length = self.parent.value('window_length')
        d = self.parent.info.dataset

        videos, begsec, endsec = d.read_videos(window_start,
                                               window_start + window_length)
        self.endsec = endsec
        medialist = vlc.MediaList(videos)
        self.medialistplayer.set_media_list(medialist)

        self.cnt_video = 0
        self.n_video = len(videos)

        self.t = QTimer()
        self.t.timeout.connect(self.check_if_finished)
        self.t.start(100)

        self.medialistplayer.play()
        self.mediaplayer.set_time(int(begsec * 1000))



from os import SEEK_SET, SEEK_END
from datetime import datetime
from struct import unpack

from numpy import fromfile

units = {-1: -1e9,  # nV
         0: -1e6,  # uV,
         1: -1e3,  # mV
         2: 1,
         100: 100,  # percent
         101: 1,  # dimentionless
         102: 1,  # dimentionless
        }

class Micromed:
    """Basic class to read the data.

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory
    """
    def __init__(self, filename):
        self.filename = filename

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header
        """
        N_ZONES = 15
        orig = {}

        with self.filename.open('rb') as f:

            f.seek(64, SEEK_SET)
            orig['surname'] = f.read(22).decode('utf-8').strip()
            orig['name'] = f.read(20).decode('utf-8').strip()
            subj_id = orig['name'] + ' ' + orig['surname']

            f.seek(128, SEEK_SET)
            day, month, year, hour, minute, sec = unpack('bbbbbb', f.read(6))
            start_time = datetime(year + 1900, month, day, hour, minute, sec)

            f.seek(138, SEEK_SET)
            BOData, n_chan, Multiplexer, s_freq, N_BYTES = unpack('IHHHH', f.read(12))

            f.seek(175, SEEK_SET)
            hdr_version = unpack('b', f.read(1))[0]
            assert hdr_version == 4

            zones = {}
            for _ in range(N_ZONES):
                zname, pos, length = unpack('8sII', f.read(16))
                zname = zname.decode('utf-8').strip()
                zones[zname] = pos, length

            pos, length = zones['ORDER']
            f.seek(pos, 0)
            order = fromfile(f, dtype='u2', count=n_chan)

            f.seek(0, SEEK_END)
            EOData = f.tell()
            n_samples = int((EOData - BOData) / (n_chan * N_BYTES))
            self.n_smp = n_samples

            chan_name =  _read_channels(f, n_chan, order, zones)[0]

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples

        TODO
        ----
        Write a reader from MICROMED data.
        """
        if begsam < 0:
            begpad = -1 * begsam
            begsam = 0
        else:
            begpad = 0

        if endsam > self.n_smp:
            endpad = endsam - self.n_smp
            endsam = self.n_smp
        else:
            endpad = 0

        return None

    def return_markers(self):
        """Return all the markers (also called triggers or events).

        Returns
        -------
        list of dict
            where each dict contains 'name' as str, 'start' and 'end' as float
            in seconds from the start of the recordings, and 'chan' as list of
            str with the channels involved (if not of relevance, it's None).

        Raises
        ------
        FileNotFoundError
            when it cannot read the events for some reason (don't use other
            exceptions).
        """
        markers = []
        return markers


def _read_channels(f, n_chan, order, zones):
    chan_names = []
    all_s_freq = []

    for c in range(n_chan):
        pos, length = zones['LABCOD']
        f.seek(pos + order[c] * 128, 0)

        chan_name = f.read(6).strip(b'\x01\x00').decode()
        chan_names.append(chan_name)
        ground = f.read(6).strip(b'\x01\x00').decode()

        logical_min, logical_max, logical_ground, physical_min, physical_max = unpack('iiiii', f.read(20))
        factor = float(physical_max - physical_min) / float( logical_max - logical_min + 1)

        k = unpack('h', f.read(2))[0]
        if k in units.keys():
            unit = units[k]
        else:
            unit = units[0]

        f.seek(8, 1)
        s_rate = unpack('H', f.read(2))[0]
        all_s_freq.append(s_rate)

        # signal = (rawdata[:, c].astype('f') - logical_ground) * factor * unit
    return chan_names,  factor, unit, all_s_freq



from datetime import datetime
from logging import getLogger

from numpy import (abs, angle, arange, around, asarray, empty, exp, linspace,
                   real, tile, zeros)
# numpy.random.random has an empty __module__ and sphinx autodoc adds it to api
from numpy import random
from numpy.fft import fft, ifft

from ..datatype import ChanTime, ChanFreq, ChanTimeFreq


lg = getLogger(__name__)


def create_data(datatype='ChanTime', start_time=None, n_trial=None,
                chan_name=None, n_chan=8, s_freq=None, time=None, freq=None,
                values=None, noise=0):
    """Create data of different datatype from scratch.

    Parameters
    ----------
    datatype : str, optional
        one of 'ChanTime', 'ChanFreq', 'ChanTimeFreq'
    start_time : datetime.datetime, optional
        starting time of the recordings
    n_trial : int, optional
        number of trials
    chan_name : list of str, optional
        names of the channels
    n_chan : int, optional
        if chan_name is not specified, this defines the number of channels
    s_freq : int, optional
        sampling frequency
    time : numpy.ndarray or tuple of two numbers, optional
        if tuple, the first and second numbers indicate beginning and end
    freq : numpy.ndarray or tuple of two numbers, optional
        if tuple, the first and second numbers indicate beginning and end
    noise : float or str, optional
        noise color to generate (white noise is 0, pink is 1, brown is 2). If
        str, then it can be 'ekg' (to create heart-related activity).
    values : tuple of two numbers, optional
        the min and max values of the random data values.

    Returns
    -------
    data : instance of specified datatype

    Notes
    -----
    ChanTime uses randn (to have normally distributed noise), while when you
    have freq, it uses random (which gives always positive values).
    You can only color noise for ChanTime, not for the other datatypes.
    """
    possible_datatypes = ('ChanTime', 'ChanFreq', 'ChanTimeFreq')
    if datatype not in possible_datatypes:
        raise ValueError('Datatype should be one of ' +
                         ', '.join(possible_datatypes))

    if n_trial is None:
        n_trial = 1

    if s_freq is None:
        s_freq = 512

    if values is None:
        values = (-1, 1)
    mult = values[1] - values[0]
    add = values[0]

    if time is not None:
        if isinstance(time, tuple) and len(time) == 2:
            time = arange(time[0], time[1], 1. / s_freq)
    else:
        time = arange(0, 1, 1. / s_freq)

    if freq is not None:
        if isinstance(freq, tuple) and len(freq) == 2:
            freq = arange(freq[0], freq[1])
    else:
        freq = arange(0, s_freq / 2. + 1)

    if chan_name is None:
        chan_name = ['chan{0:02}'.format(i) for i in range(n_chan)]

    if start_time is None:
        start_time = datetime.now()

    if datatype == 'ChanTime':
        data = ChanTime()
        data.data = empty(n_trial, dtype='O')
        for i in range(n_trial):
            if noise == 'ekg':
                n_sec = around(time[-1] - time[0])
                values = (tile(_make_ekg(s_freq), (len(chan_name), n_sec))
                          * mult + add)

            else:
                values = random.randn(*(len(chan_name), len(time))) * mult + add
                for i_ch, x in enumerate(values):
                    values[i_ch, :] = _color_noise(x, s_freq, noise)
            data.data[i] = values

    if datatype == 'ChanFreq':
        data = ChanFreq()
        data.data = empty(n_trial, dtype='O')
        for i in range(n_trial):
            data.data[i] = random.random((len(chan_name), len(freq))) * mult + add

    if datatype == 'ChanTimeFreq':
        data = ChanTimeFreq()
        data.data = empty(n_trial, dtype='O')
        for i in range(n_trial):
            data.data[i] = (random.random((len(chan_name), len(time), len(freq)))
                            * mult + add)

    data.start_time = start_time
    data.s_freq = s_freq
    data.axis['chan'] = empty(n_trial, dtype='O')
    for i in range(n_trial):
        data.axis['chan'][i] = asarray(chan_name, dtype='U')

    if datatype in ('ChanTime', 'ChanTimeFreq'):
        data.axis['time'] = empty(n_trial, dtype='O')
        for i in range(n_trial):
            data.axis['time'][i] = time

    if datatype in ('ChanFreq', 'ChanTimeFreq'):
        data.axis['freq'] = empty(n_trial, dtype='O')
        for i in range(n_trial):
            data.axis['freq'][i] = freq

    return data


def _color_noise(x, s_freq, coef=0):
    """Add some color to the noise by changing the power spectrum.

    Parameters
    ----------
    x : ndarray
        one vector of the original signal
    s_freq : int
        sampling frequency
    coef : float
        coefficient to apply (0 -> white noise, 1 -> pink, 2 -> brown,
                              -1 -> blue)

    Returns
    -------
    ndarray
        one vector of the colored noise.
    """
    # convert to freq domain
    y = fft(x)
    ph = angle(y)
    m = abs(y)

    # frequencies for each fft value
    freq = linspace(0, s_freq / 2, len(m) / 2 + 1)
    freq = freq[1:-1]

    # create new power spectrum
    m1 = zeros(len(m))
    # leave zero alone, and multiply the rest by the function
    m1[1:int(len(m) / 2)] = m[1:int(len(m) / 2)] * f(freq, coef)
    # simmetric around nyquist freq
    m1[int(len(m1) / 2 + 1):] = m1[1:int(len(m1) / 2)][::-1]

    # reconstruct the signal
    y1 = m1 * exp(1j * ph)
    return real(ifft(y1))


def f(x, coef):
    """Create an almost-linear function to apply to the power spectrum.

    Parameters
    ----------
    x : ndarray
        vector with the frequency values
    coef : float
        coefficient to apply (0 -> white noise, 1 -> pink, 2 -> brown,
                              -1 -> blue)

    Returns
    -------
    ndarray
        vector to multiply with the other frequencies

    Notes
    -----
    No activity in the frequencies below .1, to avoid huge distorsions.
    """
    y = 1 / (x ** coef)
    y[x < .1] = 0
    return y


def _make_ekg(s_freq):
    """Create a simulated EKG of one second.

    Parameters
    ----------
    s_freq : int
        sampling frequency / duration of the sample

    Returns
    -------
    ndarray
        vector of length "s_freq" with one EKG in it, with peak at 1.

    Notes
    -----
    Based on ecg.m in Matlab, in the signal toolbox.
    """
    from scipy.signal import savgol_filter  # scipy is optional dependency

    EKG_val = asarray([0, 1, 40, 1, 0, -34, 118, -99, 0, 2, 21, 2, 0, 0, 0])
    EKG_time = asarray([0, 27, 59, 91, 131, 141, 163, 185, 195, 275, 307, 339,
                        357, 390, 440, 500])

    EKG_time = around(EKG_time * s_freq / EKG_time[-2])
    EKG_time[-1] = s_freq
    x = empty(s_freq)
    for i in range(len(EKG_val) - 1):
        m = arange(EKG_time[i], EKG_time[i + 1])
        slope = (EKG_val[i + 1] - EKG_val[i]) / (EKG_time[i + 1] - EKG_time[i])
        x[EKG_time[i]:EKG_time[i + 1]] = EKG_val[i] + slope * (m - EKG_time[i])

    polyorder = int(s_freq / 40 / 2) * 2 + 1
    x = savgol_filter(x, polyorder, 0)
    return x / max(x)



"""Packages containing all the possible attributes to recordings, such as
    - channels (module "chan") with class:
        - Chan
    - anatomical info (module "anat") with class:
        - Surf
    - annotations and sleep scores (module "annotations") with class:
        - Annotations

Possibly include forward and inverse models.

These attributes are only "attached" to the DataType, there should not be any
consistency check when you load them. The risk is that attributes do not refer
to the correct datatype, but the advantage is that we cannot keep track of all
the possible inconsistencies (f.e. if the channel names are not the same
between the actual channels and those stored in the Channels class).

In addition, these classes are often used in isolation, even without a dataset,
so do not assume that any of the classes in the module can call the main
dataset. In other words, these classes shouldn't have methods calling the
datatype, but there can be functions in the modules that use both the
dataset and the classes below.

"""
from .chan import Channels
from .anat import Brain, Surf, Freesurfer
from .annotations import Annotations, create_empty_annotations



"""Package to transform the Data.

The modules in this package are classes (and maybe functions) that transform
the data into data again (all the classes defined in datatype.py). Do not
use this package to transform to other classes. If you want to transform to
basic elements, use the package "detect" for example.

"""
from .filter import filter_, convolve
from .select import select, resample
from .frequency import frequency, timefrequency
from .merge import concatenate
from .math import math
from .montage import montage
from .peaks import peaks
from .reject import rejectbadchan




"""Various functions used for the GUI.
"""
from logging import getLogger
from math import ceil, floor
from os.path import dirname, join, realpath

from PyQt5.QtCore import QRectF, QSettings, Qt
from PyQt5.QtGui import (QBrush,
                         QColor,
                         QPainterPath,
                         )
from PyQt5.QtWidgets import (QGraphicsItem,
                             QGraphicsRectItem,
                             QGraphicsSimpleTextItem,
                             QMessageBox,
                             QCommonStyle,
                             )

lg = getLogger(__name__)

LINE_WIDTH = 0  # COSMETIC LINE
LINE_COLOR = 'black'


MAX_LENGTH = 20

stdicon = QCommonStyle.standardIcon


icon_path = join(dirname(realpath(__file__)), 'icons')
oxy_path = join(icon_path, 'oxygen')

ICON = {'open_rec': join(oxy_path, 'document-open.png'),
        'page_prev': join(oxy_path, 'go-previous-view.png'),
        'page_next': join(oxy_path, 'go-next-view.png'),
        'step_prev': join(oxy_path, 'go-previous.png'),
        'step_next': join(oxy_path, 'go-next.png'),
        'chronometer': join(oxy_path, 'chronometer.png'),
        'up': join(oxy_path, 'go-up.png'),
        'down': join(oxy_path, 'go-down.png'),
        'zoomin': join(oxy_path, 'zoom-in.png'),
        'zoomout': join(oxy_path, 'zoom-out.png'),
        'zoomnext': join(oxy_path, 'zoom-next.png'),
        'zoomprev': join(oxy_path, 'zoom-previous.png'),
        'ydist_more': join(oxy_path, 'format-line-spacing-triple.png'),
        'ydist_less': join(oxy_path, 'format-line-spacing-normal.png'),
        'selchan': join(oxy_path, 'mail-mark-task.png'),
        'widget': join(oxy_path, 'window-duplicate.png'),
        'settings': join(oxy_path, 'configure.png'),
        'quit': join(oxy_path, 'window-close.png'),
        'bookmark': join(oxy_path, 'bookmarks-organize.png'),
        'event': join(oxy_path, 'edit-table-cell-merge.png'),
        'new_eventtype': join(oxy_path, 'edit-table-insert-column-right.png'),
        'del_eventtype': join(oxy_path, 'edit-table-delete-column.png'),
        }

settings = QSettings("wonambi", "wonambi")


class Path(QPainterPath):
    """Paint a line in the simplest possible way.

    Parameters
    ----------
    x : ndarray or list
        x-coordinates
    y : ndarray or list
        y-coordinates
    """
    def __init__(self, x, y):
        super().__init__()

        self.moveTo(x[0], y[0])
        for i_x, i_y in zip(x, y):
            self.lineTo(i_x, i_y)


class RectMarker(QGraphicsRectItem):
    """Class to draw a rectangular, coloured item.
    
    Parameters
    ----------
    x : float
        x position in scene
    y : gloat
        y position in scene
    width : float
        length in seconds
    height : float
        height in scene units
    color : str or QColor, optional
        color of the rectangle
    """
    def __init__(self, x, y, width, height, zvalue, color='blue'):
        super().__init__()
        
        self.color = color
        self.setZValue(zvalue)
        buffer = 1
        self.marker = QRectF(x, y, width, height)
        self.b_rect = QRectF(x - buffer/2, y + buffer/2, width + buffer,
                             height + buffer)

    def boundingRect(self):
        return self.b_rect
        
    def paint(self, painter, option, widget):
        color = QColor(self.color)
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.marker)
        super().paint(painter, option, widget)


class TextItem_with_BG(QGraphicsSimpleTextItem):
    """Class to draw text with dark background (easier to read).

    Parameters
    ----------
    bg_color : str or QColor, optional
        color to use as background
    """
    def __init__(self, bg_color='black'):
        super().__init__()

        self.bg_color = bg_color
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations)
        self.setBrush(QBrush(Qt.white))

    def paint(self, painter, option, widget):
        bg_color = QColor(self.bg_color)
        painter.setBrush(QBrush(bg_color))
        painter.drawRect(self.boundingRect())
        super().paint(painter, option, widget)


def keep_recent_datasets(max_dataset_history, info=None):
    """Keep track of the most recent recordings.

    Parameters
    ----------
    max_dataset_history : int
        maximum number of datasets to remember
    info : str, optional TODO
        path to file

    Returns
    -------
    list of str
        paths to most recent datasets (only if you don't specify
        new_dataset)
    """
    history = settings.value('recent_recordings', [])
    if isinstance(history, str):
        history = [history]

    if info is not None and info.filename is not None:
        if info.repo is not None:
            # this will parsed by self.info.open_dataset()
            new_dataset = info.filename + ' (' + info.repo + ')'
        else:
            new_dataset = info.filename

        if new_dataset in history:
            lg.debug(new_dataset + ' already present, will be replaced')
            history.remove(new_dataset)
        if len(history) > max_dataset_history:
            lg.debug('Removing last dataset ' + history[-1])
            history.pop()

        lg.info('Adding ' + new_dataset + ' to list of recent datasets')
        history.insert(0, new_dataset)
        settings.setValue('recent_recordings', history)
        return None

    else:
        return history


def choose_file_or_dir():
    """Create a simple message box to see if the user wants to open dir or file

    Returns
    -------
    str
        'dir' or 'file' or 'ieegorg' or 'abort'

    """
    question = QMessageBox(QMessageBox.Information, 'Open Dataset',
                           'Do you want to open a file, a directory or a '
                           'remote ieeg.org dataset?')
    dir_button = question.addButton('Directory', QMessageBox.YesRole)
    file_button = question.addButton('File', QMessageBox.NoRole)
    remote_button = question.addButton('Remote Dataset', QMessageBox.NoRole)
    question.addButton(QMessageBox.Cancel)
    question.exec_()
    response = question.clickedButton()

    if response == dir_button:
        return 'dir'
    elif response == file_button:
        return 'file'
    elif response == remote_button:
        return 'remote'
    else:
        return 'abort'


def short_strings(s, max_length=MAX_LENGTH):
    if len(s) > max_length:
        max_length -= 3  # dots
        start = ceil(max_length / 2)
        end = -floor(max_length / 2)
        s = s[:start] + '...' + s[end:]
    return s


def convert_name_to_color(s):
    """Convert any string to an RGB color.

    Parameters
    ----------
    s : str
        string to convert
    selection : bool, optional
        if an event is being selected, it's lighter

    Returns
    -------
    instance of QColor
        one of the possible color

    Notes
    -----
    It takes any string and converts it to RGB color. The same string always
    returns the same color. The numbers are a bit arbitrary but not completely.
    h is the baseline color (keep it high to have brighter colors). Make sure
    that the max module + h is less than 256 (RGB limit).

    The number you multiply ord for is necessary to differentiate the letters
    (otherwise 'r' and 's' are too close to each other).
    """
    h = 100
    v = [5 * ord(x) for x in s]
    sum_mod = lambda x: sum(x) % 100
    color = QColor(sum_mod(v[::3]) + h, sum_mod(v[1::3]) + h,
                   sum_mod(v[2::3]) + h)
    return color



"""Functions used when creating a new window.

"""
from logging import getLogger
from functools import partial

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QAction,
                             QDockWidget,
                             QMessageBox,
                             )

from .settings import (Settings, SlowWaveHelp, SpindleHelp,
                       EvtAnalysisHelp)  # has to be first
from .utils import ICON  # has to be second
from .labels import Labels
from .channels import Channels
from .info import Info
from .overview import Overview
from .notes import (Notes, MergeDialog, SpindleDialog, SWDialog,
                    EventAnalysisDialog)
from .analysis import AnalysisDialog
from .spectrum import Spectrum
from .traces import Traces
from .video import Video


lg = getLogger(__name__)


def create_widgets(MAIN):
    """Create all the widgets and dockwidgets. It also creates actions to
    toggle views of dockwidgets in dockwidgets.
    """

    """ ------ CREATE WIDGETS ------ """
    MAIN.labels = Labels(MAIN)
    MAIN.channels = Channels(MAIN)
    MAIN.notes = Notes(MAIN)
    MAIN.merge_dialog = MergeDialog(MAIN)
    MAIN.spindle_dialog = SpindleDialog(MAIN)
    MAIN.slow_wave_dialog = SWDialog(MAIN)
    MAIN.spindle_help = SpindleHelp(MAIN)
    MAIN.slowwave_help = SlowWaveHelp(MAIN)
    MAIN.event_analysis_dialog = EventAnalysisDialog(MAIN)
    MAIN.evt_analysis_help = EvtAnalysisHelp(MAIN)
    MAIN.analysis_dialog = AnalysisDialog(MAIN)
    MAIN.overview = Overview(MAIN)
    MAIN.spectrum = Spectrum(MAIN)
    MAIN.traces = Traces(MAIN)
    MAIN.video = Video(MAIN)
    MAIN.settings = Settings(MAIN)  # depends on all widgets apart from Info
    MAIN.info = Info(MAIN)  # this has to be the last, it depends on settings

    MAIN.setCentralWidget(MAIN.traces)

    """ ------ LIST DOCKWIDGETS ------ """
    new_docks = [{'name': 'Information',
                  'widget': MAIN.info,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Labels',
                  'widget': MAIN.labels,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Channels',
                  'widget': MAIN.channels,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Spectrum',
                  'widget': MAIN.spectrum,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Annotations',
                  'widget': MAIN.notes,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Video',
                  'widget': MAIN.video,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Overview',
                  'widget': MAIN.overview,
                  'main_area': Qt.BottomDockWidgetArea,
                  'extra_area': Qt.TopDockWidgetArea,
                  },
                 ]

    """ ------ CREATE DOCKWIDGETS ------ """
    idx_docks = {}
    actions = MAIN.action

    actions['dockwidgets'] = []
    for dock in new_docks:
        dockwidget = QDockWidget(dock['name'], MAIN)
        dockwidget.setWidget(dock['widget'])
        dockwidget.setAllowedAreas(dock['main_area'] | dock['extra_area'])
        dockwidget.setObjectName(dock['name'])  # savestate

        idx_docks[dock['name']] = dockwidget
        MAIN.addDockWidget(dock['main_area'], dockwidget)

        dockwidget_action = dockwidget.toggleViewAction()
        dockwidget_action.setIcon(QIcon(ICON['widget']))

        actions['dockwidgets'].append(dockwidget_action)

    """ ------ ORGANIZE DOCKWIDGETS ------ """
    MAIN.tabifyDockWidget(idx_docks['Information'],
                          idx_docks['Video'])
    MAIN.tabifyDockWidget(idx_docks['Channels'],
                          idx_docks['Labels'])
    idx_docks['Information'].raise_()


def create_actions(MAIN):
    """Create all the possible actions."""
    actions = MAIN.action  # actions was already taken

    """ ------ OPEN SETTINGS ------ """
    actions['open_settings'] = QAction(QIcon(ICON['settings']), 'Settings',
                                       MAIN)
    actions['open_settings'].triggered.connect(MAIN.show_settings)

    """ ------ CLOSE WINDOW ------ """
    actions['close_wndw'] = QAction(QIcon(ICON['quit']), 'Quit', MAIN)
    actions['close_wndw'].triggered.connect(MAIN.close)

    """ ------ ABOUT ------ """
    actions['about'] = QAction('About WONAMBI', MAIN)
    actions['about'].triggered.connect(MAIN.about)

    actions['aboutqt'] = QAction('About Qt', MAIN)
    actions['aboutqt'].triggered.connect(lambda: QMessageBox.aboutQt(MAIN))


def create_menubar(MAIN):
    """Create the whole menubar, based on actions."""
    actions = MAIN.action
    menubar = MAIN.menuBar()
    menubar.clear()

    """ ------ FILE ------ """
    menu_file = menubar.addMenu('File')
    menu_file.addAction(MAIN.info.action['open_dataset'])
    submenu_recent = menu_file.addMenu('Recent Datasets')
    submenu_recent.addActions(MAIN.info.action['open_recent'])

    menu_file.addSeparator()
    menu_file.addAction(actions['open_settings'])
    menu_file.addSeparator()
    menu_file.addAction(actions['close_wndw'])

    """ ------ CHANNELS ------ """
    actions = MAIN.channels.action
    menu_time = menubar.addMenu('Channels')
    menu_time.addAction(actions['load_channels'])
    menu_time.addAction(actions['save_channels'])

    """ ------ NAVIGATION ------ """
    actions = MAIN.traces.action

    menu_time = menubar.addMenu('Navigation')
    menu_time.addAction(actions['step_prev'])
    menu_time.addAction(actions['step_next'])
    menu_time.addAction(actions['page_prev'])
    menu_time.addAction(actions['page_next'])
    menu_time.addSeparator()
    menu_time.addAction(actions['addtime_-6h'])
    menu_time.addAction(actions['addtime_-1h'])
    menu_time.addAction(actions['addtime_-10min'])
    menu_time.addAction(actions['addtime_10min'])
    menu_time.addAction(actions['addtime_1h'])
    menu_time.addAction(actions['addtime_6h'])
    menu_time.addSeparator()
    menu_time.addAction(actions['go_to_epoch'])
    menu_time.addAction(actions['line_up_with_epoch'])

    """ ------ VIEW ------ """
    actions = MAIN.traces.action

    menu_view = menubar.addMenu('View')
    submenu_ampl = menu_view.addMenu('Global Scaling')
    submenu_ampl.addAction(actions['Y_less'])
    submenu_ampl.addAction(actions['Y_more'])
    submenu_ampl.addSeparator()
    for x in sorted(MAIN.value('y_scale_presets'), reverse=True):
        act = submenu_ampl.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.Y_ampl, x))

    submenu_dist = menu_view.addMenu('Distance Between Traces')
    submenu_dist.addAction(actions['Y_wider'])
    submenu_dist.addAction(actions['Y_tighter'])
    submenu_dist.addSeparator()
    for x in sorted(MAIN.value('y_distance_presets'),
                    reverse=True):
        act = submenu_dist.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.Y_dist, x))

    submenu_length = menu_view.addMenu('Window Length')
    submenu_length.addAction(actions['X_more'])
    submenu_length.addAction(actions['X_less'])
    submenu_length.addSeparator()
    for x in sorted(MAIN.value('window_length_presets'),
                    reverse=True):
        act = submenu_length.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.X_length, x))

    menu_view.addAction(actions['cross_chan_mrk'])

    """ ------ ANNOTATIONS ------ """
    actions = MAIN.notes.action

    menu_annot = menubar.addMenu('Annotations')
    menu_annot.addAction(actions['new_annot'])
    menu_annot.addAction(actions['load_annot'])
    menu_annot.addAction(actions['clear_annot'])
    menu_annot.addSeparator()

    submenu_rater = menu_annot.addMenu('Rater')
    submenu_rater.addAction(actions['new_rater'])
    submenu_rater.addAction(actions['del_rater'])
    submenu_rater.addSeparator()
    if MAIN.notes.annot is not None:
        for rater in sorted(MAIN.notes.annot.raters):
            act = submenu_rater.addAction(rater)
            act.triggered.connect(partial(MAIN.notes.select_rater, rater))
    menu_annot.addSeparator()

    submenu_marker = menu_annot.addMenu('Bookmark')
    submenu_marker.addAction(actions['new_bookmark'])

    submenu_event = menu_annot.addMenu('Event')
    submenu_event.addAction(actions['new_eventtype'])
    submenu_event.addAction(actions['del_eventtype'])
    submenu_event.addAction(actions['merge_events'])

    # these are the real QActions attached to notes
    submenu_stage = menu_annot.addMenu('Stage')
    submenu_stage.addActions(MAIN.notes.actions())

    submenu_mrkr = menu_annot.addMenu('Cycle')
    submenu_mrkr.addAction(actions['cyc_start'])
    submenu_mrkr.addAction(actions['cyc_end'])
    submenu_mrkr.addAction(actions['remove_cyc'])
    submenu_mrkr.addAction(actions['clear_cyc'])
    menu_annot.addSeparator()

    submenu_import = menu_annot.addMenu('Import staging')
    submenu_import.addAction(actions['import_alice'])
    submenu_import.addAction(actions['import_compumedics'])
    submenu_import.addAction(actions['import_domino'])
    submenu_import.addAction(actions['import_remlogic'])
    submenu_import.addAction(actions['import_sandman'])
    submenu_import.addAction(actions['import_fasst'])

    menu_annot.addAction(actions['export'])

    """ ------ ANALYSIS ------ """
    actions = MAIN.notes.action

    menu_analysis = menubar.addMenu('Analysis')

    submenu_detect = menu_analysis.addMenu('Detection')
    submenu_detect.addAction(actions['spindle'])
    submenu_detect.addAction(actions['slow_wave'])

    menu_analysis.addAction(actions['analyze_events'])    
    menu_analysis.addAction(actions['analyze'])


    """ ------ WINDOWS ------ """
    actions = MAIN.action

    menu_window = menubar.addMenu('Windows')
    for dockwidget_act in actions['dockwidgets']:
        menu_window.addAction(dockwidget_act)
    MAIN.menu_window = menu_window

    menu_about = menubar.addMenu('About')
    menu_about.addAction(actions['about'])
    menu_about.addAction(actions['aboutqt'])

def create_toolbar(MAIN):
    """Create the various toolbars."""
    actions = MAIN.action

    toolbar = MAIN.addToolBar('File Management')
    toolbar.setObjectName('File Management')  # for savestate
    toolbar.addAction(MAIN.info.action['open_dataset'])
    toolbar.addSeparator()
    toolbar.addAction(MAIN.channels.action['load_channels'])
    toolbar.addAction(MAIN.channels.action['save_channels'])
    toolbar.addSeparator()
    toolbar.addAction(MAIN.notes.action['new_annot'])
    toolbar.addAction(MAIN.notes.action['load_annot'])

    """ ------ SCROLL ------ """
    actions = MAIN.traces.action

    toolbar = MAIN.addToolBar('Scroll')
    toolbar.setObjectName('Scroll')  # for savestate
    toolbar.addAction(actions['step_prev'])
    toolbar.addAction(actions['step_next'])
    toolbar.addAction(actions['page_prev'])
    toolbar.addAction(actions['page_next'])
    toolbar.addSeparator()
    toolbar.addAction(actions['X_more'])
    toolbar.addAction(actions['X_less'])
    toolbar.addSeparator()
    toolbar.addAction(actions['Y_less'])
    toolbar.addAction(actions['Y_more'])
    toolbar.addAction(actions['Y_wider'])
    toolbar.addAction(actions['Y_tighter'])

    """ ------ ANNOTATIONS ------ """
    actions = MAIN.notes.action

    toolbar = MAIN.addToolBar('Annotations')
    toolbar.setObjectName('Annotations')

    toolbar.addAction(actions['new_bookmark'])
    toolbar.addSeparator()
    toolbar.addAction(actions['new_event'])
    toolbar.addWidget(MAIN.notes.idx_eventtype)
    toolbar.addSeparator()
    toolbar.addWidget(MAIN.notes.idx_stage)
    toolbar.addWidget(MAIN.notes.idx_quality)



"""Large and simple widget to indicate settings/Settings.
"""
from ast import literal_eval
from logging import getLogger

from PyQt5.QtCore import QSettings, Qt
from PyQt5.QtWidgets import (QCheckBox,
                             QComboBox,
                             QDialogButtonBox,
                             QDialog,
                             QFileDialog,
                             QFormLayout,
                             QGroupBox,
                             QHBoxLayout,
                             QLineEdit,
                             QListWidget,
                             QPushButton,
                             QSplitter,
                             QStackedWidget,
                             QTextEdit,
                             QVBoxLayout,
                             QWidget,
                             )

lg = getLogger(__name__)

settings = QSettings("wonambi", "scroll_data")


# DO NOT DUPLICATE NAMES
DEFAULTS = {}
DEFAULTS['channels'] = {'hp': .5,
                        'lp': 45,
                        'color': 'black',
                        'scale': 1,
                        }
DEFAULTS['overview'] = {'timestamp_steps': 60 * 60,
                        'overview_scale': 30,
                        }
DEFAULTS['spectrum'] = {'x_min': 0.,
                        'x_max': 30.,
                        'x_tick': 10.,
                        'y_min': -5.,
                        'y_max': 5.,
                        'y_tick': 5.,
                        'log': True,
                        }
DEFAULTS['notes'] = {'marker_show': True,
                     'marker_color': 'darkBlue',
                     'annot_show': True,
                     'annot_bookmark_color': 'darkMagenta',
                     'min_marker_dur': .1,
                     'scoring_window': 30,
                     }
DEFAULTS['traces'] = {'n_time_labels': 3,
                      'y_distance': 50.,
                      'y_scale': 1.,
                      'label_ratio': 0.05,
                      'max_s_freq': 30000,
                      'window_start': 0,
                      'window_length': 30,
                      'window_step': 5,
                      'grid_x': True,
                      'grid_xtick': 1,  # in seconds
                      'grid_y': True,
                      'grid_ytick': 35,
                      }
DEFAULTS['settings'] = {'max_dataset_history': 20,
                        'y_distance_presets': [20., 30., 40., 50., 100., 200.],
                        'y_scale_presets': [.1, .2, .5, 1, 2, 5, 10],
                        'window_length_presets': [1., 5., 10., 20., 30., 60.],
                        'recording_dir': '/home/gio/recordings',
                        }
DEFAULTS['video'] = {}


class Settings(QDialog):
    """Window showing the Settings/settings.

    Parameters
    ----------
    parent : instance of QMainWindow
        the main window
    """
    def __init__(self, parent):
        super().__init__(None, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
        self.parent = parent
        self.config = ConfigUtils(self.parent.refresh)

        self.setWindowTitle('Settings')
        self.create_settings()

    def create_settings(self):
        """Create the widget, organized in two parts.

        Notes
        -----
        When you add widgets in config, remember to update show_settings too
        """
        bbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Apply |
                                QDialogButtonBox.Cancel)
        self.idx_ok = bbox.button(QDialogButtonBox.Ok)
        self.idx_apply = bbox.button(QDialogButtonBox.Apply)
        self.idx_cancel = bbox.button(QDialogButtonBox.Cancel)
        bbox.clicked.connect(self.button_clicked)

        page_list = QListWidget()
        page_list.setSpacing(1)
        page_list.currentRowChanged.connect(self.change_widget)

        pages = ['General', 'Overview', 'Signals', 'Channels', 'Spectrum',
                 'Notes', 'Video']
        for one_page in pages:
            page_list.addItem(one_page)

        self.stacked = QStackedWidget()
        self.stacked.addWidget(self.config)
        self.stacked.addWidget(self.parent.overview.config)
        self.stacked.addWidget(self.parent.traces.config)
        self.stacked.addWidget(self.parent.channels.config)
        self.stacked.addWidget(self.parent.spectrum.config)
        self.stacked.addWidget(self.parent.notes.config)
        self.stacked.addWidget(self.parent.video.config)

        hsplitter = QSplitter()
        hsplitter.addWidget(page_list)
        hsplitter.addWidget(self.stacked)

        btnlayout = QHBoxLayout()
        btnlayout.addStretch(1)
        btnlayout.addWidget(bbox)

        vlayout = QVBoxLayout()
        vlayout.addWidget(hsplitter)
        vlayout.addLayout(btnlayout)

        self.setLayout(vlayout)

    def change_widget(self, new_row):
        """Change the widget on the right side.

        Parameters
        ----------
        new_row : int
            index of the widgets
        """
        self.stacked.setCurrentIndex(new_row)

    def button_clicked(self, button):
        """Action when button was clicked.

        Parameters
        ----------
        button : instance of QPushButton
            which button was pressed
        """
        if button in (self.idx_ok, self.idx_apply):

            # loop over widgets, to see if they were modified
            for i_config in range(self.stacked.count()):
                one_config = self.stacked.widget(i_config)

                if one_config.modified:
                    lg.debug('Settings for ' + one_config.widget +
                             ' were modified')
                    one_config.get_values()

                    if self.parent.info.dataset is not None:
                        one_config.update_widget()
                    one_config.modified = False

            if button == self.idx_ok:
                self.accept()

        if button == self.idx_cancel:
            self.reject()


class Config(QWidget):
    """Base class for widgets used in the Settings.

    Parameters
    ----------
    widget : str
        name of the widget
    update_widget : function
        function to run to update the main window with new values

    Attributes
    ----------
    modified : bool
        if the preference widget has been changed
    value : dict
        dictionary with the actual current values
    index : dict
        dictionary with the instances of the small widgets

    Notes
    -----
    You'll need to implement create_config with the QGroupBox and layouts
    """
    def __init__(self, widget, update_widget):
        super().__init__()

        self.modified = False
        self.widget = widget

        value_names = list(DEFAULTS[widget].keys())
        self.value = self.create_values(value_names)
        self.index = self.create_indices(value_names)

        self.create_config()
        self.put_values()
        self.update_widget = update_widget

    def create_values(self, value_names):
        """Read original values from the settings or the defaults.

        Parameters
        ----------
        value_names : list of str
            list of value names to read

        Returns
        -------
        dict
            dictionary with the value names as keys
        """
        output = {}
        for value_name in value_names:
            output[value_name] = read_settings(self.widget, value_name)

        return output

    def create_indices(self, value_names):
        """Create empty indices as None. They'll be created by create_config.

        """
        return dict(zip(value_names, [None] * len(value_names)))

    def get_values(self):
        """Get values from the GUI and save them in preference file."""
        for value_name, widget in self.index.items():
            self.value[value_name] = widget.get_value(self.value[value_name])

            setting_name = self.widget + '/' + value_name
            settings.setValue(setting_name, self.value[value_name])

    def put_values(self):
        """Put values to the GUI.

        Notes
        -----
        In addition, when one small widget has been changed, it calls
        set_modified, so that we know that the preference widget was modified.

        """
        for value_name, widget in self.index.items():
            widget.set_value(self.value[value_name])
            widget.connect(self.set_modified)

    def create_config(self):
        """Placeholder: it'll be replaced with actual layout."""
        pass

    def set_modified(self):
        """Simply mark that the preference widget was modified.

        Notes
        -----
        You cannot use lambda because they don't accept assignments.

        """
        self.modified = True


class ConfigUtils(Config):

    def __init__(self, update_widget):
        super().__init__('settings', update_widget)

    def create_config(self):

        box0 = QGroupBox('History')
        self.index['max_dataset_history'] = FormInt()
        self.index['recording_dir'] = FormStr()

        form_layout = QFormLayout()
        form_layout.addRow('Max History Size',
                           self.index['max_dataset_history'])
        form_layout.addRow('Directory with recordings',
                           self.index['recording_dir'])
        box0.setLayout(form_layout)

        box1 = QGroupBox('Default values')
        self.index['y_distance_presets'] = FormList()
        self.index['y_scale_presets'] = FormList()
        self.index['window_length_presets'] = FormList()

        form_layout = QFormLayout()
        form_layout.addRow('Signal scaling, presets',
                           self.index['y_scale_presets'])
        form_layout.addRow('Distance between signals, presets',
                           self.index['y_distance_presets'])
        form_layout.addRow('Window length, presets',
                           self.index['window_length_presets'])
        box1.setLayout(form_layout)

        main_layout = QVBoxLayout()
        main_layout.addWidget(box0)
        main_layout.addWidget(box1)
        main_layout.addStretch(1)

        self.setLayout(main_layout)


def read_settings(widget, value_name):
    """Read Settings information, either from INI or from default values.

    Parameters
    ----------
    widget : str
        name of the widget
    value_name : str
        name of the value of interest.

    Returns
    -------
    multiple types
        type depends on the type in the default values.

    """
    setting_name = widget + '/' + value_name
    default_value = DEFAULTS[widget][value_name]

    default_type = type(default_value)
    if default_type is list:
        default_type = type(default_value[0])

    val = settings.value(setting_name, default_value, type=default_type)
    return val


class HelpDialog(QDialog):
    """Generic help dialog, showing uneditable HTML text."""
    def __init__(self):
        super().__init__(None, Qt.WindowSystemMenuHint | Qt.WindowTitleHint)
        self.setWindowTitle('Help')
        self.setWindowModality(Qt.ApplicationModal)

        self.create_dialog()

    def create_dialog(self):
        bbox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.idx_ok = bbox.button(QDialogButtonBox.Ok)
        bbox.clicked.connect(self.accept)

        textbox = QTextEdit(self)
        textbox.setReadOnly(True)
        self.textbox = textbox

        btnlayout = QHBoxLayout()
        btnlayout.addStretch(1)
        btnlayout.addWidget(bbox)

        vlayout = QVBoxLayout()
        vlayout.addWidget(textbox)
        vlayout.addLayout(btnlayout)

        self.setLayout(vlayout)

    def set_text(self, message):
        """Insert HTML text."""
        self.textbox.insertHtml(message)


class SpindleHelp(HelpDialog):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        s = ('<h3>Spindle detection help</h3>'
             ''
             'Note: Epochs marked as bad signal are excluded from detection.'
             '<br /><br />'
             '<b>Label</b><br />'
             'Label for this event type, to appear on trace.<br /><br />'
             '<b>Channel group</b><br />'
             'Channel group containing desired channel(s).<br /><br />'
             '<b>Channel(s)</b><br />'
             'One or several channels on which to detect slow waves.<br />'
             '<br />'
             '<b>Stage(s)</b><br />'
             'One or several stages during which to detect slow waves. '
             'If none are selected, all stages will be scanned.<br /><br />'
             '<b>Merge events across channels</b>'
             'If selected, spindles detected on different channels and '
             'separated by less than <b>Minimum interval</b> will be merged '
             'into a single event, to the channel with the earliest onset '
             'spindle.<br />'
             ''
             '<h4>Spindle detection methods</h4>'
             '<p>'
             'The method options in the spindle detection dialog are direct '
             'implementations of spindle detection methods '
             'reported in the scholarly articles cited below. '
             'For customizable parameters (in bold), the values used in the '
             'article are specified in square brackets []. '
             'Further details on the original method are provided in italics. '
             '</p>'

             '<p>'
             '<b>Wamsley, E. J. et al. (2012) Biol. Psychiatry 71, 154-61</b>'
             '<ol>'
             '<i><li>Authors analyzed spindles during NREM2. EEG data were '
             'filtered at 0.5–35Hz.</li></i>'
             '<li>The artifact-free EEG signal is subjected to a time-'
             'frequency transformation using an 8-parameter complex Morlet '
             'wavelet with the average of <b>Lowcut</b> and <b>Highcut</b> as '
             'the frequency, with σ = <b>Wavelet sigma</b> [0.5] s and with '
             'window size = <b>Detection window</b> [1] s.</li>'
             '<li>The moving average of the rectified signal is calculated, '
             'using a sliding window of size = <b>Smoothing</b> [0.1] s.</li>'
             '<li>A spindle event is identified whenever this wavelet signal '
             'exceeds threshold, defined as <b>Detection threshold</b> [4.5] '
             'times the mean signal amplitude of all artifact-free '
             'epochs, between <b>Minimum duration</b> [0.4] s '
             'and <b>Maximum duration</b> s [no maximum]. In this '
             'implementation, threshold crossings define the spindle start and '
             'end times, but see next point for the original method.</li>'
             '<i><li>The duration of each spindle was calculated as the '
             'half-height width of wavelet energy within the spindle '
             'frequency range.</li></i>'
             '</ol>'
             '</p>'

             '<p>'
             '<b>Nir, Y. et al. (2011) Neuron 70, 153-69</b>'
             '<ol>'
             '<i><li>The channels with spindle activity in NREM sleep are '
             'chosen for further analysis (significant spectral power '
             'increases in spindle range as compared with a 1/f model, '
             'p &lsaquo; 0.001, paired t-test across 10 s segments.)</li></i>'
             '<li>The EEG signal is bandpass filtered between <b>Lowcut</b> '
             '[10] Hz and <b>Highcut</b> [16] Hz with a zero-phase 4th order '
             'Butterworth filter. </li>'
             '<li>Instantaneous amplitude in the sigma frequency is '
             'extracted via the Hilbert transform.</li>'
             '<li>To avoid excessive multiple crossings of thresholds within '
             'the same spindle event, instantaneous amplitude is temporally '
             'smoothed using a Gaussian kernel of σ = <b>Smoothing</b> [0.4] '
             's.'
             '</li>'
             '<li>Events with amplitude greater than mean + <b>Detection '
             'threshold</b> [3] SD (computed across all artifact-free <i>NREM '
             'sleep</i> epochs) are considered putative spindles and '
             'detections within <b>Minimum interval</b> [1] s are merged.</li>'
             '<li>A threshold of mean + <b>Selection threshold</b> [1] SD '
             'defines start and end times, and events with duration between '
             '<b>Minimum duration</b> [0.5] s and <b>Maximum duration</b> [2] '
             's are selected for further analysis.'
             '</i></li>'
             '<i><li>Those channels, in which an increase in spectral power '
             'within the detected events was restricted to the spindle-'
             'frequency range (10-16 Hz) rather than broadband (unpaired '
             't-test (α=0.001) between maximal spectral power in detected vs. '
             'random events), and with at least 1 spindle per min of NREM '
             'sleep were chosen for further analysis. Authors focused on the '
             'frontal and parietal channels, where spindle-events could be '
             'reliably detected and successfully separated from paroxysmal '
             'events. This highly conservative procedure of including in the '
             'analysis only the channels with high spindle SNR, ensured that '
             'local occurrence of spindle events does not arise merely as a '
             'result of the lack of spindles or poor spindle SNR in some '
             'channels.</li></i>'
             '</ol>'
             '</p>'

             '<p>'
             '<b>Mölle, M. et al. (2011) Sleep 34, 1411-21</b>'
             '<ol>'
             '<li>The <i>NREM</i> signal is bandpass filtered between '
             '<b>Lowcut</b> and <b>Highcut</b>, using a zero-phase 4th order '
             'Butterworth filter. '
             '<i>Authors used the adapted sigma peak +/- 1.5 Hz.</i></li>'
             '<li>The root-mean-square of the signal is taken, with a '
             'moving window of size = <b>Detection window</b> [0.2] s.</li>'
             '<li>The resulting RMS signal is smoothed with a moving '
             'average of window size = <b>Smoothing</b> [0.2] s.</li>'
             '<li>Spindles are detected as a continuous rise in the '
             'smoothed RMS signal above <b>Detection threshold</b> [1.5] SDs, '
             'lasting between <b>Minimum duration</b> [0.5] s and <b>Maximum '
             'duration</b> [3] s. Spindle start and end times are the '
             'threshold crossings.</li>'
             '</ol>'
             '</p>'

             '<p>'
             '<b>Ferrarelli, F. et al. (2007) Am. J. Psychiatry 164, 483-92'
             '</b>'
             '<ol>'
             '<li>The <i>NREM</i> signal is bandpass filtered between '
             '<b>Lowcut</b> [12] Hz and <b>Highcut</b> [15] Hz. Authors did '
             'not specify a filter; this software employs a zero-phase 4th '
             'order Butterworth filter.</li>'
             '<li>The filtered signal is rectified.</li>'
             '<li>Upper and lower thresholds are set at <b>Detection '
             'threshold</b> [8] and <b>Selection threshold</b> [2] times the '
             'average amplitude of the time series.</li>'
             '<li>Spindles are detected when the time series amplitude '
             'exceeds the upper threshold. This is the spindle peak.</li>'
             '<li>Spindle start and end times are set at the time points '
             'immediately preceding and following this peak when the '
             'amplitude of the time series drops below the lower threshold.'
             '</li>'
             '<li>The authors did not set minimum or maximum durations.</li>'
             '</ol>'
             '</p>'

             '<p>'
             '<b>UCSD</b>'
             '<ol>'
             '<li>The raw EEG signal is subjected to a time-frequency '
             'transformation using real wavelets with frequencies from '
             '<b>Lowcut</b> to <b>Highcut</b> at 0.5-Hz intervals, with width '
             '= 0.5 s and with window size = <b>Detection window</b> [1] s.'
             '</li>'
             '<li>The resulting time-frequency signals are rectified and '
             'convolved with a Tukey window of size = 0.5 s, then '
             'averaged to produce a single time-frequency signal.</li>'
             '<li>A threshold is defined as the signal median plus '
             '<b>Detection threshold</b> [2] SDs.</li>'
             '<li>Spindles are detected at each relative maximum in the '
             'signal which exceeds the threshold.</li>'
             '<li>Steps 1-3 are repeated on the raw signal, this time with '
             'width = 0.2 s, with Tukey window size = 0.2 s, and with the '
             'threshold set at <b>Selection threshold</b> [1] SD.</li>'
             '<li>Spindle start and end times are defined at threshold '
             'crossings.</li>'
             '<li>Spindles are retained if their duration is between '
             '<b>Minimum duration</b> and <b>Maximum duration</b>.</li>'
             ''
             '</ol></p>'
             )
        self.resize(500, 700)
        self.set_text(s)


class SlowWaveHelp(HelpDialog):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        s = ('<h3>Slow wave detection help</h3>'
             ''
             'Note: Epochs marked as bad signal are excluded from detection.'
             '<br /><br />'
             '<b>Label</b><br />'
             'Label for this event type, to appear on trace.<br /><br />'
             '<b>Channel group</b><br />'
             'Channel group containing desired channel(s).<br /><br />'
             '<b>Channel(s)</b><br />'
             'One or several channels on which to detect slow waves.<br />'
             '<br />'
             '<b>Stage(s)</b><br />'
             'One or several stages during which to detect slow waves. '
             'If none are selected, all stages will be scanned.<br /><br />'
             '<b>Invert detection</b><br />'
             'If selected, the algorithm will detect inverted slow waves, '
             'i.e. positive-then-negative instead of negative-then-positive. '
             '<br />'
             ''
             '<h4>Slow wave detection methods</h4>'
             '<p>'
             'The method options in the slow wave detection dialog are direct '
             'implementations of slow wave detection methods '
             'reported in the scholarly articles cited below. '
             'For customizable parameters (in bold), the values used in the '
             'article are specified in square brackets []. '
             'Further details on the original method are provided in italics. '
             '</p>'

             '<p>'
             '<b>Massimini, M. et al. (2004) J Neurosci 24(31), 6862-70</b>'
             '<ol>'
             '<i><li>256-channel EEG is re-referenced to the average of the '
             'signals from the earlobes.</li></i>'
             '<i><li>EEG signal is locally averaged over 4 non-overlapping '
             'regions of the scalp.</li></i>'
             '<li>The <i>NREM</i> signal is bandpass filtered between '
             '<b>Lowcut</b> and <b>Highcut</b>, using a zero-phase 4th order '
             'Butterworth filter. '
             '<li>Slow waves are detected when the following 3 criteria are '
             'met:'
             '<ol>'
             '<li>A negative zero crossing and a subsequent positive zero '
             'crossing separated by <b>Minimum trough duration</b> [0.3] '
             'and <b>Maximum trough duration</b> [1.0] s.</li>'
             '<li>A negative peak between the two zero crossings with voltage '
             'less than <b>Maximum trough amplitude</b> [-80] &mu;V'
             '<li>A negative-to-positive peak-to-peak amplitude greater than '
             '<b>Minimum peak-to-peak amplitude</b> [140] &mu;V.'
             '</ol></ol>'
             '</p>'
             ''
             '<p>'
             '<b>AASM/Massimini2004</b><br /><br />'
             'This is a reimplementation of Massimini et al., 2004 (above), '
             'except with default values for slow waves as defined by the '
             'American Academy of Sleep Medicine (AASM).'
             )
        self.resize(500, 700)
        self.set_text(s)


class EvtAnalysisHelp(HelpDialog):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        s = ('<h2>Event analysis</h2>'
             ''
             '<h3>Bandpass</h3>'
             '<h4>Lowcut and Highcut</h4>'
             'Lower and upper band limits for analysis. For <b>Maximum '
             'amplitude</b>, <b>Peak-to-peak amplitude</b> and <b>RMS</b>, '
             'limits are used to bandpass filter the signal, using a 4th-'
             'order zero-phase Butterworth filter. '
             'For <b>Peak frequency</b> and <b>Average power</b>, limits set '
             'the minimum and maximum frequencies of interest.'
             '<br />'
             ''
             '<h3>Parameters, global</h3>'
             '<h4>Count</h4>'
             'Total number of events detected in the queried stage(s).'
             '<h4>Density</h4>'
             'Number of events per epoch of the queried stage(s). '
             'Epoch duration can be set in Preferences &rarr; Notes &rarr; '
             'Length of scoring window. Change will not take effect until a '
             'new rater is created.'
             '<br />'
             ''
             '<h3>Parameters, per event</h3>'
             '<h4>Duration</h4>'
             'Duration of the event in seconds. Event start and end are as '
             'seen on the trace.'
             '<h4>Maximum amplitude</h4>'
             'Maximum value of the event signal, in &mu;V, after bandpass '
             'filtering. If raw signal is inverted, this will give the signal '
             'minimum (most negative value).'
             '<h4>Peak-to-peak amplitude</h4>'
             'Absolute difference between the event signal maximum and '
             'minimum, in &mu;V, after bandpass filtering.'
             '<h4>Peak frequency</h4>'
             'Frequency, in Hz, associated with the highest power spectral '
             'density within the specified band, after removal of 1/f noise '
             '(signal whitening). Power spectral density is estimated with a '
             'simple periodogram.'
             '<h4>Average power</h4>'
             'Average value of the power spectral density within the '
             'specified band, in &mu;V&sup2;, after  removal of 1/f noise '
             '(signal whitening). Power spectral density is estimated with a '
             'simple periodogram.'
             '<h4>RMS</h4>'
             'Square root of the mean of the square of all event signal '
             'values, in &mu;V, after (1) bandpass filtering and (2) removal '
             'of 1/f noise (signal whitening). '
             '<br />'
             ''
             '<h3>Options</h3>'
             '<h4>Log transform</h4>'
             'If checked, all parameters (global and per event) will be '
             'reported as their natural logarithms. This allows the use of '
             'general linear model staistics on log-normally distributed '
             'data. '
             'For more information, see <i>Buzsáki G, Mizuseki K. The '
             'log-dynamic brain: how skewed distributions affect network '
             'operations. Nature reviews. Neuroscience. 2014 Apr;15(4):264.'
             '</i>'
             '<h4>Frequency split</h4>'
             'If checked, events will be analyzed separately based on whether '
             'their peak frequency is above or below the specified frequency. '
             'Global parameters and averages will be calculated separately '
             'for both groups, and output will be saved in separate CSV '
             'files.'
             '<h4>Cycle split</h4>'
             'If checked, events will be analyzed separately by sleep cycle. '
             'Global parameters and averages will be calculated separately '
             'for each group, and output will be saved in separate CSV files.'
             )
        self.resize(500, 700)
        self.set_text(s)


class FormBool(QCheckBox):
    """Subclass QCheckBox to have a more consistent API across widgets.

    Parameters
    ----------
    checkbox_label : str
        label next to checkbox

    """
    def __init__(self, checkbox_label):
        super().__init__(checkbox_label)

    def get_value(self, default=False):
        """Get the value of the QCheckBox, as boolean.

        Parameters
        ----------
        default : bool
            not used

        Returns
        -------
        bool
            state of the checkbox

        """
        return self.checkState() == Qt.Checked

    def set_value(self, value):
        """Set value of the checkbox.

        Parameters
        ----------
        value : bool
            value for the checkbox

        """
        if value:
            self.setCheckState(Qt.Checked)
        else:
            self.setCheckState(Qt.Unchecked)

    def connect(self, funct):
        """Call funct when user ticks the box.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.stateChanged.connect(funct)


class FormFloat(QLineEdit):
    """Subclass QLineEdit for float to have a more consistent API across
    widgets.

    """
    def __init__(self, default=None):
        super().__init__('')
        
        if default is not None:
            self.set_value(default)

    def get_value(self, default=0):
        """Get float from widget.

        Parameters
        ----------
        default : float
            default value for the parameter in case it fails

        Returns
        -------
        float
            the value in text or default

        """
        text = self.text()
        try:
            text = float(text)
        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to float.' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)
        return text

    def set_value(self, value):
        """Set value of the float.

        Parameters
        ----------
        value : float
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormInt(QLineEdit):
    """Subclass QLineEdit for int to have a more consistent API across widgets.

    """
    def __init__(self, default=None):
        super().__init__('')
        
        if default is not None:
            self.set_value(default)

    def get_value(self, default=0):
        """Get int from widget.

        Parameters
        ----------
        default : int
            default value for the parameter in case it fails

        Returns
        -------
        int
            the value in text or default

        """
        text = self.text()
        try:
            text = int(float(text))  # to convert values like 30.0
        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to int. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)
        return text

    def set_value(self, value):
        """Set value of the int.

        Parameters
        ----------
        value : int
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormList(QLineEdit):
    """Subclass QLineEdit for lists to have a more consistent API across
    widgets.

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=None):
        """Get int from widget.

        Parameters
        ----------
        default : list
            list with widgets

        Returns
        -------
        list
            list that might contain int or str or float etc

        """
        if default is None:
            default = []

        try:
            text = literal_eval(self.text())
            if not isinstance(text, list):
                pass
                # raise ValueError

        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to list. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)

        return text

    def set_value(self, value):
        """Set value of the list.

        Parameters
        ----------
        value : list
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormStr(QLineEdit):
    """Subclass QLineEdit for str to have a more consistent API across widgets.

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=''):
        """Get int from widget.

        Parameters
        ----------
        default : str
            not used

        Returns
        -------
        str
            the value in text

        """
        return self.text()

    def set_value(self, value):
        """Set value of the string.

        Parameters
        ----------
        value : str
            value for the line edit

        """
        self.setText(value)

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormDir(QPushButton):
    """Subclass QPushButton for str to have a more consistent API across widgets.

    Notes
    -----
    It calls to open the directory three times, but I don't understand why

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=''):
        """Get int from widget.

        Parameters
        ----------
        default : str
            not used

        Returns
        -------
        str
            the value in text

        """
        return self.text()

    def set_value(self, value):
        """Set value of the string.

        Parameters
        ----------
        value : str
            value for the line edit

        """
        self.setText(value)

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        Notes
        -----
        There is something wrong here. When you run this function, it calls
        for opening a directory three or four times. This is obviously wrong
        but I don't understand why this happens three times. Traceback did not
        help.

        """
        def get_directory():
            rec = QFileDialog.getExistingDirectory(self,
                                                   'Path to Recording'
                                                   ' Directory')
            if rec == '':
                return

            self.setText(rec)
            funct()

        self.clicked.connect(get_directory)

class FormMenu(QComboBox):
    """Subclass QComboBox for dropdown menus to have a more consistent API 
    across widgets.
    
    Parameters
    ----------
    input_list: list of str
        items to include in the dropdown menu / combobox
    """
    def __init__(self, input_list):
        super().__init__()
        
        if input_list is not None:
            for i in input_list:
                self.addItem(i)

    def get_value(self, default=None):
        """Get selection from widget.

        Parameters
        ----------
        default : str
            str for use by widget

        Returns
        -------
        str
            selected item from the combobox

        """
        if default is None:
            default = ''

        try:
            text = self.currentText()

        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to list. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)

        return text

    def set_value(self, value):
        """Set value of the list.

        Parameters
        ----------
        value : str
            value for the combobox

        """
        self.setCurrentText(str(value))

    def connect(self, funct):
        """Call funct when the selection was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.currentIndexChanged.connect(funct)


from numpy import isnan

from wonambi import Dataset
from wonambi.ioeeg import write_edf
from wonambi.utils import create_data

from .paths import (psg_file,
                    generated_file,
                    EXPORTED_PATH,
                    )


psg = Dataset(psg_file)
generated = Dataset(generated_file)


def test_edf_read():
    psg.read_data(begtime=10, endtime=20)
    markers = psg.read_markers()
    assert len(markers) == 0


def test_edf_before_start_both():
    data = psg.read_data(begsam=-100, endsam=-10)
    assert isnan(data.data[0][0, 0])


def test_edf_before_start():
    data = psg.read_data(begsam=-100, endsam=10)
    assert isnan(data.data[0][0, 0])


def test_edf_after_end():
    n_samples = psg.header['n_samples']
    data = psg.read_data(begsam=n_samples - 100, endsam=n_samples + 100)
    assert isnan(data.data[0][0, -1])


def test_edf_annot():
    markers = generated.read_markers()
    assert len(markers) == 2


def test_edf_write():
    data = create_data()
    write_edf(data, EXPORTED_PATH / 'export.edf')



"""Widget to define channels, montage and filters.
"""
from copy import deepcopy
from json import dump, load
from os.path import splitext

from PyQt5.QtGui import (QColor,
                         )

from PyQt5.QtWidgets import (QAbstractItemView,
                             QAction,
                             QColorDialog,
                             QDoubleSpinBox,
                             QFileDialog,
                             QFormLayout,
                             QGridLayout,
                             QGroupBox,
                             QHBoxLayout,
                             QInputDialog,
                             QLabel,
                             QListWidget,
                             QListWidgetItem,
                             QPushButton,
                             QVBoxLayout,
                             QTabWidget,
                             QWidget
                             )


from .settings import Config, FormFloat, FormStr


class ConfigChannels(Config):
    """Widget with preferences in Settings window for Channels."""
    def __init__(self, update_widget):
        super().__init__('channels', update_widget)

    def create_config(self):
        box0 = QGroupBox('Channels')

        self.index['hp'] = FormFloat()
        self.index['lp'] = FormFloat()
        self.index['color'] = FormStr()
        self.index['scale'] = FormFloat()

        form_layout = QFormLayout()
        form_layout.addRow('Default High-Pass Filter', self.index['hp'])
        form_layout.addRow('Default Low-Pass Filter', self.index['lp'])
        form_layout.addRow('Default Color', self.index['color'])
        form_layout.addRow('Default Scale', self.index['scale'])
        box0.setLayout(form_layout)

        main_layout = QVBoxLayout()
        main_layout.addWidget(box0)
        main_layout.addStretch(1)

        self.setLayout(main_layout)


class ChannelsGroup(QWidget):
    """Tab inside the Channels widget.

    Parameters
    ----------
    chan_name : list of str
        list of all the channels in the dataset
    config_value : dict
        default values for the channels
    s_freq : int
        sampling frequency (to define max of filter)

    Attributes
    ----------
    chan_name : list of str
        list of all the channels in the dataset

    idx_l0 : QListWidget
        list with the channels to plot
    idx_l1 : QListWidget
        list with the channels to use as reference
    idx_hp : QDoubleSpinBox
        spin box to indicate the high-pass filter
    idx_lp : QDoubleSpinBox
        spin box to indicate the low-pass filter
    idx_scale : QDoubleSpinBox
        spin_box to indicate the group-specific scaling
    idx_reref : QPushButton
        it triggers a selection of reference channels equal to the channels to
        plot.
    idx_color : QColor
        color of the traces beloning to this channel group (it could be a
        property of QWidget)

    Notes
    -----
    TODO: re-referencing should be more flexible, by allowing other types of
    referencing.

    Use config_value instead of config, because it's easier to pass dict
    when loading channels montage.
    """
    def __init__(self, chan_name, group_name, config_value, s_freq):
        super().__init__()

        self.chan_name = chan_name
        self.group_name = group_name

        self.idx_l0 = QListWidget()
        self.idx_l1 = QListWidget()

        self.add_channels_to_list(self.idx_l0)
        self.add_channels_to_list(self.idx_l1)

        self.idx_hp = QDoubleSpinBox()
        self.idx_hp.setValue(config_value['hp'])
        self.idx_hp.setSuffix(' Hz')
        self.idx_hp.setDecimals(1)
        self.idx_hp.setMaximum(s_freq / 2)
        self.idx_hp.setToolTip('0 means no filter')

        self.idx_lp = QDoubleSpinBox()
        self.idx_lp.setValue(config_value['lp'])
        self.idx_lp.setSuffix(' Hz')
        self.idx_lp.setDecimals(1)
        self.idx_lp.setMaximum(s_freq / 2)
        self.idx_lp.setToolTip('0 means no filter')

        self.idx_scale = QDoubleSpinBox()
        self.idx_scale.setValue(config_value['scale'])
        self.idx_scale.setSuffix('x')

        self.idx_reref = QPushButton('Average')
        self.idx_reref.clicked.connect(self.rereference)

        self.idx_color = QColor(config_value['color'])

        l_form = QFormLayout()
        l_form.addRow('High-Pass', self.idx_hp)
        l_form.addRow('Low-Pass', self.idx_lp)

        r_form = QFormLayout()
        r_form.addRow('Scaling', self.idx_scale)
        r_form.addRow('Reference', self.idx_reref)

        l0_layout = QVBoxLayout()
        l0_layout.addWidget(QLabel('Active'))
        l0_layout.addWidget(self.idx_l0)

        l1_layout = QVBoxLayout()
        l1_layout.addWidget(QLabel('Reference'))
        l1_layout.addWidget(self.idx_l1)

        l_layout = QHBoxLayout()
        l_layout.addLayout(l0_layout)
        l_layout.addLayout(l1_layout)

        lr_form = QHBoxLayout()
        lr_form.addLayout(l_form)
        lr_form.addLayout(r_form)

        layout = QVBoxLayout()
        layout.addLayout(l_layout)
        layout.addLayout(lr_form)

        self.setLayout(layout)

    def add_channels_to_list(self, l):
        """Create list of channels (one for those to plot, one for ref).

        Parameters
        ----------
        l : instance of QListWidget
            one of the two lists (chan_to_plot or ref_chan)
        """
        l.clear()

        l.setSelectionMode(QAbstractItemView.ExtendedSelection)
        for chan in self.chan_name:
            item = QListWidgetItem(chan)
            l.addItem(item)

    def highlight_channels(self, l, selected_chan):
        """Highlight channels in the list of channels.

        Parameters
        ----------
        selected_chan : list of str
            channels to indicate as selected.
        """
        for row in range(l.count()):
            item = l.item(row)
            if item.text() in selected_chan:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def rereference(self):
        """Automatically highlight channels to use as reference, based on
        selected channels."""
        selectedItems = self.idx_l0.selectedItems()

        chan_to_plot = []
        for selected in selectedItems:
            chan_to_plot.append(selected.text())
        self.highlight_channels(self.idx_l1, chan_to_plot)

    def get_info(self):
        """Get the information about the channel groups.

        Returns
        -------
        dict
            information about this channel group

        Notes
        -----
        The items in selectedItems() are ordered based on the user's selection
        (which appears pretty random). It's more consistent to use the same
        order of the main channel list. That's why the additional for-loop
        is necessary. We don't care about the order of the reference channels.
        """
        selectedItems = self.idx_l0.selectedItems()
        selected_chan = [x.text() for x in selectedItems]
        chan_to_plot = []
        for chan in self.chan_name:
            if chan in selected_chan:
                chan_to_plot.append(chan)

        selectedItems = self.idx_l1.selectedItems()
        ref_chan = []
        for selected in selectedItems:
            ref_chan.append(selected.text())

        hp = self.idx_hp.value()
        if hp == 0:
            low_cut = None
        else:
            low_cut = hp

        lp = self.idx_lp.value()
        if lp == 0:
            high_cut = None
        else:
            high_cut = lp

        scale = self.idx_scale.value()

        group_info = {'name': self.group_name,
                      'chan_to_plot': chan_to_plot,
                      'ref_chan': ref_chan,
                      'hp': low_cut,
                      'lp': high_cut,
                      'scale': float(scale),
                      'color': self.idx_color
                      }

        return group_info


class Channels(QWidget):
    """Widget with information about channel groups.

    Attributes
    ----------
    parent : QMainWindow
        the main window
    config : ConfigChannels
        preferences for this widget

    filename : path to file
        file with the channel groups
    groups : list of dict
        each dict contains information about one channel group

    tabs : QTabWidget
        Widget that contains the tabs with channel groups
    """
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.config = ConfigChannels(lambda: None)

        self.filename = None
        self.groups = []

        self.tabs = None

        self.create()
        self.create_action()

    def create(self):
        """Create Channels Widget"""
        add_button = QPushButton('New')
        add_button.clicked.connect(self.new_group)
        color_button = QPushButton('Color')
        color_button.clicked.connect(self.color_group)
        del_button = QPushButton('Delete')
        del_button.clicked.connect(self.del_group)
        apply_button = QPushButton('Apply')
        apply_button.clicked.connect(self.apply)

        self.button_add = add_button
        self.button_color = color_button
        self.button_del = del_button
        self.button_apply = apply_button

        buttons = QGridLayout()
        buttons.addWidget(add_button, 0, 0)
        buttons.addWidget(color_button, 1, 0)
        buttons.addWidget(del_button, 0, 1)
        buttons.addWidget(apply_button, 1, 1)

        self.tabs = QTabWidget()

        layout = QVBoxLayout()
        layout.addLayout(buttons)
        layout.addWidget(self.tabs)

        self.setLayout(layout)

        self.setEnabled(False)
        self.button_color.setEnabled(False)
        self.button_del.setEnabled(False)
        self.button_apply.setEnabled(False)

    def create_action(self):
        """Create actions related to channel selection."""
        actions = {}

        act = QAction('Load Montage...', self)
        act.triggered.connect(self.load_channels)
        act.setEnabled(False)
        actions['load_channels'] = act

        act = QAction('Save Montage...', self)
        act.triggered.connect(self.save_channels)
        act.setEnabled(False)
        actions['save_channels'] = act

        self.action = actions

    def update(self):
        self.setEnabled(True)
        self.action['load_channels'].setEnabled(True)
        self.action['save_channels'].setEnabled(True)

    def new_group(self, checked=False, test_name=None):
        """Create a new channel group.

        Parameters
        ----------
        checked : bool
            comes from QAbstractButton.clicked
        test_name : str
            used for testing purposes to avoid modal window

        Notes
        -----
        Don't call self.apply() just yet, only if the user wants it.
        """
        chan_name = self.parent.labels.chan_name
        if chan_name is None:
            self.parent.statusBar().showMessage('No dataset loaded')

        else:
            if test_name is None:
                new_name = QInputDialog.getText(self, 'New Channel Group',
                                                'Enter Name')
            else:
                new_name = [test_name, True]  # like output of getText

            if new_name[1]:
                s_freq = self.parent.info.dataset.header['s_freq']
                group = ChannelsGroup(chan_name, new_name[0],
                                      self.config.value, s_freq)
                self.tabs.addTab(group, new_name[0])
                self.tabs.setCurrentIndex(self.tabs.currentIndex() + 1)

                # activate buttons
                self.button_color.setEnabled(True)
                self.button_del.setEnabled(True)
                self.button_apply.setEnabled(True)

    def color_group(self, checked=False, test_color=None):
        """Change the color of the group."""
        group = self.tabs.currentWidget()
        if test_color is None:
            newcolor = QColorDialog.getColor(group.idx_color)
        else:
            newcolor = test_color
        group.idx_color = newcolor

        self.apply()

    def del_group(self):
        """Delete current group."""
        idx = self.tabs.currentIndex()
        self.tabs.removeTab(idx)

        self.apply()

    def apply(self):
        """Apply changes to the plots."""
        self.read_group_info()

        if self.tabs.count() == 0:
            # disactivate buttons
            self.button_color.setEnabled(False)
            self.button_del.setEnabled(False)
            self.button_apply.setEnabled(False)
        else:
            # activate buttons
            self.button_color.setEnabled(True)
            self.button_del.setEnabled(True)
            self.button_apply.setEnabled(True)

        if self.groups:
            self.parent.overview.update_position()
            self.parent.spectrum.update()
        else:
            self.parent.traces.reset()
            self.parent.spectrum.reset()

    def read_group_info(self):
        """Get information about groups directly from the widget."""
        self.groups = []
        for i in range(self.tabs.count()):
            one_group = self.tabs.widget(i).get_info()
            # one_group['name'] = self.tabs.tabText(i)
            self.groups.append(one_group)

    def load_channels(self, checked=False, test_name=None):
        """Load channel groups from file.

        Parameters
        ----------
        test_name : path to file
            when debugging the function, you can open a channels file from the
            command line
        """
        chan_name = self.parent.labels.chan_name

        if self.filename is not None:
            filename = self.filename
        elif self.parent.info.filename is not None:
            filename = (splitext(self.parent.info.filename)[0] +
                        '_channels.json')
        else:
            filename = None

        if test_name is None:
            filename, _ = QFileDialog.getOpenFileName(self,
                                                      'Open Channels Montage',
                                                      filename,
                                                      'Channels File (*.json)')
        else:
            filename = test_name

        if filename == '':
            return

        self.filename = filename
        with open(filename, 'r') as outfile:
            groups = load(outfile)

        s_freq = self.parent.info.dataset.header['s_freq']
        no_in_dataset = []
        for one_grp in groups:
            no_in_dataset.extend(set(one_grp['chan_to_plot']) -
                                 set(chan_name))
            chan_to_plot = set(chan_name) & set(one_grp['chan_to_plot'])
            ref_chan = set(chan_name) & set(one_grp['ref_chan'])

            group = ChannelsGroup(chan_name, one_grp['name'], one_grp, s_freq)
            group.highlight_channels(group.idx_l0, chan_to_plot)
            group.highlight_channels(group.idx_l1, ref_chan)
            self.tabs.addTab(group, one_grp['name'])

        if no_in_dataset:
            msg = 'Channels not present in the dataset: '
            self.parent.statusBar().showMessage(msg + ', '.join(no_in_dataset))

        self.apply()

    def save_channels(self, checked=False, test_name=None):
        """Save channel groups to file."""
        self.read_group_info()

        if self.filename is not None:
            filename = self.filename
        elif self.parent.info.filename is not None:
            filename = (splitext(self.parent.info.filename)[0] +
                        '_channels.json')
        else:
            filename = None

        if test_name is None:
            filename, _ = QFileDialog.getSaveFileName(self,
                                                      'Save Channels Montage',
                                                      filename,
                                                      'Channels File (*.json)')
        else:
            filename = test_name

        if filename == '':
            return

        self.filename = filename

        groups = deepcopy(self.groups)
        for one_grp in groups:
            one_grp['color'] = one_grp['color'].rgba()

        with open(filename, 'w') as outfile:
            dump(groups, outfile, indent=' ')

    def reset(self):
        """Reset all the information of this widget."""
        self.filename = None
        self.groups = []

        self.tabs.clear()

        self.setEnabled(False)
        self.button_color.setEnabled(False)
        self.button_del.setEnabled(False)
        self.button_apply.setEnabled(False)
        self.action['load_channels'].setEnabled(False)
        self.action['save_channels'].setEnabled(False)



from logging import getLogger
from numpy import nan, nanargmax, nanargmin

lg = getLogger(__name__)


def peaks(data, method='max', axis='time', limits=None):
    """Return the values of an index where the data is at max or min

    Parameters
    ----------
    method : str, optional
        'max' or 'min'
    axis : str, optional
        the axis where you want to detect the peaks
    limits : tuple of two values, optional
        the lowest and highest limits where to search for the peaks
    data : instance of Data
        one of the datatypes

    Returns
    -------
    instance of Data
        with one dimension less that the input data. The actual values in
        the data can be not-numberic, for example, if you look for the
        max value across electrodes

    Notes
    -----
    This function is useful when you want to find the frequency value at which
    the power is the largest, or to find the time point at which the signal is
    largest, or the channel at which the activity is largest.
    """
    idx_axis = data.index_of(axis)
    output = data._copy()
    output.axis.pop(axis)

    for trl in range(data.number_of('trial')):
        values = data.axis[axis][trl]
        dat = data(trial=trl)

        if limits is not None:
            limits = (values < limits[0]) | (values > limits[1])

            idx = [slice(None)] * len(data.list_of_axes)
            idx[idx_axis] = limits
            dat[idx] = nan

        if method == 'max':
            peak_val = nanargmax(dat, axis=idx_axis)
        elif method == 'min':
            peak_val = nanargmin(dat, axis=idx_axis)

        output.data[trl] = values[peak_val]

    return output



"""explain sess behavior"""

from datetime import datetime
from urllib.parse import urljoin
from base64 import standard_b64encode
from getpass import getpass
import hashlib
from xml.etree import ElementTree

from numpy import array, fromstring, newaxis

try:
    import requests
except ImportError as import_err:
    requests = import_err  # import_err is local to except-statement

HOST = 'www.ieeg.org'
HTTP = 'https://'
URL_ID = '/services/timeseries/getIdByDataSnapshotName/'
URL_DETAILS = '/services/timeseries/getDataSnapshotTimeSeriesDetails/'
URL_DATA = '/services/timeseries/getUnscaledTimeSeriesSetBinaryRaw/'


SESS = None


class IEEG_org:

    def __init__(self, name):
        global SESS
        if SESS is None:
            SESS = Session()
        self._sess = SESS

        self.filename = name
        self._snapshot = self._sess.get_snapshot(name)
        xml = self._sess.get_dataset(self._snapshot)
        self._details = ElementTree.fromstring(xml).find('details')

    def return_hdr(self):
        chan_name = [chan.find('channelLabel').text for chan in self._details]
        factor = [float(chan.find('voltageConversionFactor').text) for chan in self._details]
        self._factor = array(factor)

        one_chan = self._details[0]
        subj_id = self.filename
        start_time = datetime.fromtimestamp(int(one_chan.find('startTime').text) / 1e6)
        s_freq = float(one_chan.find('sampleRate').text)
        self.s_freq = s_freq
        n_samples = int(one_chan.find('numberOfSamples').text)
        orig = self._details

        self._chan_id = array([chan.find('revisionId').text for chan in self._details])

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """order of chan is taken into account by ieeg.org server"""
        start = int(begsam / self.s_freq * 1000000)
        duration = int((endsam - begsam) / self.s_freq * 1000000)
        params = {'start': start, 'duration': duration}

        xml_str = _prepare_xml_str(self._chan_id[chan])
        r = self._sess.get_data(self._snapshot, params, xml_str)

        h = r.headers
        samples_per_row = set(h['samples-per-row'].split(','))
        if not len(samples_per_row) == 1:
            raise ValueError('Not all channels have equal length')
        conv = array([float(x) for x in h['voltage-conversion-factors-mv'].split(',')])

        dat = fromstring(r.content, dtype='>i4')
        n_smp = int(list(samples_per_row)[0])
        # TODO: should we use self._factor
        dat = dat.reshape(-1, n_smp) * conv[:, newaxis]

        return dat

    def return_markers(self):
        return []


class Session:
    def __init__(self, username=None, password=None, password_md5=None):
        if username is None:
            username = input('ieeg.org username:')
        self.username = username

        if password is None and password_md5 is None:
            password = getpass('ieeg.org password:')

        if password:
            self.password = md5(password)
        elif password_md5:
            self.password = password_md5

    def get_snapshot(self, name):
        return self._get_info(URL_ID + name)

    def get_dataset(self, snapshot):
        return self._get_info(URL_DETAILS + snapshot)

    def get_data(self, snapshot, params, xml_str):

        path = URL_DATA + snapshot
        dtime, sig = self._make_signature('POST', path, params, xml_str)
        headers = {'username': self.username,
                   'timestamp': dtime,
                   'signature': sig,
                   'Content-Type': 'application/xml'}
        url = urljoin(HTTP + HOST, path)
        r = requests.post(url, headers=headers, params=params, data=xml_str)
        return r

    def _get_info(self, path):
        dtime, sig = self._make_signature('GET', path)
        headers = {'username': self.username,
                   'timestamp': dtime,
                   'signature': sig,
                   'Content-Type': 'application/xml'}
        url = urljoin(HTTP + HOST, path)

        if isinstance(requests, BaseException):
            raise requests  # import error

        r = requests.get(url, headers=headers)
        return r.content.decode()

    def _make_signature(self, method, urlpath, params=None, xml_str=''):
        if params is None:
            params = {}
        dtime = datetime.now().isoformat()

        params_str = '&'.join(k + '=' + str(v) for k, v in params.items())

        xml_str = _sha256(xml_str)
        to_hash = (self.username, self.password, method, HOST, urlpath,
                   params_str, dtime, xml_str)
        return dtime, _sha256('\n'.join(to_hash))


def _prepare_xml_str(chan_id):
    XML_BEG = ('<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
               '<timeSeriesIdAndDChecks><timeSeriesIdAndDChecks>')
    XML_END = '</timeSeriesIdAndDChecks></timeSeriesIdAndDChecks>'
    XML_CHAN = ('<timeSeriesIdAndCheck><dataCheck>{}</dataCheck>'
                '<id>{}</id></timeSeriesIdAndCheck>')

    s = XML_BEG
    for i_chan in chan_id:
        s = s + XML_CHAN.format(i_chan, i_chan)
    s = s + XML_END

    return s


def md5(s):
    """ Returns MD5 hashed string """
    m = hashlib.md5(s.encode())
    return m.hexdigest()


def _sha256(s):
    m = hashlib.sha256(s.encode())
    return standard_b64encode(m.digest()).decode()



"""Package to import and export OpBox Physiology format.
"""
from datetime import datetime, timedelta
from pathlib import Path
from struct import unpack
from os import SEEK_END

from numpy import c_, empty, float64, NaN, memmap


N_HDR_BYTES = 12
BYTESIZE = float64().itemsize


class OpBox:
    """Class to read the data in opbox format, which is fast to write and read

    Parameters
    ----------
    filename : path to file
        the name of the filename with extension .phy
    """
    def __init__(self, filename):
        self.filename = filename

        self._n_samples = None
        self._n_chan_in_dat = None

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : empty list
            no additional info present
        """
        with open(self.filename, 'rb') as f:
            s_freq, n_anlg, n_dgtl = unpack('<iii', f.read(12))
            f.seek(0, SEEK_END)
            eof = f.tell()
        s_freq = float(s_freq)

        p = Path(self.filename)
        try:
            subj_id, date, time = p.stem.split('-')
            start_time = datetime.strptime(date + time, '%Y%m%d%H%M%S')
        except ValueError:
            # if it can't parse info from name, use arbitrary info
            subj_id = ''
            start_time = datetime(1900, 1, 1, 0, 0, 0)

        n_chan = n_anlg + n_dgtl
        n_chan_in_dat = n_chan + 1  # add timestamp

        n_samples = int((eof - N_HDR_BYTES) / n_chan_in_dat / BYTESIZE)

        chan_name = ['a{:03d}'.format(x) for x in range(n_anlg)]
        chan_name.extend(['d{:03d}'.format(x) for x in range(n_dgtl)])

        orig = []  # no additional info

        self._n_samples = n_samples
        self._n_chan_in_dat = n_chan_in_dat

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples. To save memory, the
            data are memory-mapped, and you cannot change the values on disk.
        """
        if isinstance(chan, int):  # make sure it's a list
            chan = [chan, ]

        data = memmap(self.filename, dtype='float64', mode='c',
                      shape=(self._n_chan_in_dat, self._n_samples), order='F',
                      offset=N_HDR_BYTES)
        data = data[1:, :]  # ignore timestamps

        begrec = max((begsam, 0))
        endrec = min((endsam, self._n_samples))
        dat = data[chan, begrec:endrec].astype(float64)

        if begsam < 0:
            pad = empty((dat.shape[0], 0 - begsam))
            pad.fill(NaN)
            dat = c_[pad, dat]

        if endsam >= self._n_samples:

            pad = empty((dat.shape[0], endsam - self._n_samples))
            pad.fill(NaN)
            dat = c_[dat, pad]

        return dat

    def return_markers(self):
        """This format doesn't have markers.

        Returns
        -------
        empty list

        Raises
        ------
        FileNotFoundError
            when it cannot read the events for some reason (don't use other
            exceptions).
        """
        return []



"""Module contains all the exceptions

"""

class UnrecognizedFormat(Exception):
    """Could not recognize the format of the file for channels.

    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



#!/usr/bin/env python3

try:
    from PyQt5.QtWidgets import QApplication
except ImportError:
    raise ImportError('You need to install PyQt5 to run GUI')


from logging import getLogger, INFO, StreamHandler, Formatter

lg = getLogger('wonambi')
FORMAT = '%(asctime)s %(filename)s/%(funcName)s (%(levelname)s): %(message)s'
DATE_FORMAT = '%H:%M:%S'

formatter = Formatter(fmt=FORMAT, datefmt=DATE_FORMAT)
handler = StreamHandler()
handler.setFormatter(formatter)

lg.handlers = []
lg.addHandler(handler)

lg.setLevel(INFO)

from datetime import datetime
now = datetime.now()
from types import MethodType

from PyQt5.QtCore import QSettings
from PyQt5.QtWidgets import QMainWindow, QMessageBox

from . import __version__
from .widgets.creation import (create_menubar, create_toolbar,
                               create_actions, create_widgets)
from .widgets.settings import DEFAULTS
from .widgets.utils import keep_recent_datasets

settings = QSettings("wonambi", "wonambi")
lg.info('Reading settings from {}'.format(settings.fileName()))


class MainWindow(QMainWindow):
    """Create an instance of the main window.

    """
    def __init__(self):
        super().__init__()

        self.info = None
        self.labels = None
        self.channels = None
        self.spectrum = None
        self.overview = None
        self.notes = None
        self.video = None
        self.traces = None
        self.settings = None

        self.action = {}  # actions was already taken
        self.menu_window = None

        # I prefer to have these functions in a separate module, for clarify
        self.create_widgets = MethodType(create_widgets, self)
        self.create_actions = MethodType(create_actions, self)
        self.create_menubar = MethodType(create_menubar, self)
        self.create_toolbar = MethodType(create_toolbar, self)

        self.create_widgets()
        self.create_actions()
        self.create_menubar()
        self.create_toolbar()

        self.statusBar()
        self.setWindowTitle('WONAMBI v' + __version__)

        window_geometry = settings.value('window/geometry')
        if window_geometry is not None:
            self.restoreGeometry(window_geometry)
        window_state = settings.value('window/state')
        if window_state is not None:
            self.restoreState(window_state)

        self.show()

    def refresh(self):
        """Functions to re-run once settings have been changed."""
        self.create_menubar()

    def value(self, parameter, new_value=None):
        """This function is a shortcut for any parameter. Instead of calling
        the widget, its config and its values, you can call directly the
        parameter.

        Parameters
        ----------
        parameter : str
            name of the parameter of interest
        new_value : str or float, optional
            new value for the parameter

        Returns
        -------
        str or float
            if you didn't specify new_value, it returns the current value.

        Notes
        -----
        It's important to maintain an organized dict in DEFAULTS which has to
        correspond to the values in the widgets, also the name of the widget.
        DEFAULTS is used like a look-up table.
        """
        for widget_name, values in DEFAULTS.items():
            if parameter in values.keys():
                widget = getattr(self, widget_name)
                if new_value is None:
                    return widget.config.value[parameter]
                else:
                    lg.debug('setting value {0} of {1} to {2}'
                             ''.format(parameter, widget_name, new_value))
                    widget.config.value[parameter] = new_value

    def update(self):
        """Once you open a dataset, it activates all the widgets.
        """
        self.info.display_dataset()
        self.overview.update()
        self.labels.update(labels=self.info.dataset.header['chan_name'])
        self.channels.update()

        try:
            self.info.markers = self.info.dataset.read_markers()
        except FileNotFoundError:
            lg.info('No notes/markers present in the header of the file')
        else:
            self.notes.update_dataset_marker()

    def reset(self):
        """Remove all the information from previous dataset before loading a
        new dataset.
        """

        # store current dataset
        max_dataset_history = self.value('max_dataset_history')
        keep_recent_datasets(max_dataset_history, self.info)

        # reset all the widgets
        self.labels.reset()
        self.channels.reset()
        self.info.reset()
        self.notes.reset()
        self.overview.reset()
        self.spectrum.reset()
        self.traces.reset()

    def show_settings(self):
        """Open the Setting windows, after updating the values in GUI. """
        self.notes.config.put_values()
        self.overview.config.put_values()
        self.settings.config.put_values()
        self.spectrum.config.put_values()
        self.traces.config.put_values()
        self.video.config.put_values()

        self.settings.show()

    def show_merge_dialog(self):
        """Create the event merging dialog."""
        self.merge_dialog.update_event_types()
        self.merge_dialog.show()

    def show_spindle_dialog(self):
        """Create the spindle detection dialog."""
        self.spindle_dialog.update_groups()
        self.spindle_dialog.show()

    def show_slow_wave_dialog(self):
        """Create the SW detection dialog."""
        self.slow_wave_dialog.update_groups()
        self.slow_wave_dialog.show()

    def show_event_analysis_dialog(self):
        """Create the event analysis dialog."""
        self.event_analysis_dialog.update_types()
        self.event_analysis_dialog.update_groups()
        self.event_analysis_dialog.update_cycles()
        self.event_analysis_dialog.show()

    def show_analysis_dialog(self):
        """Create the analysis dialog."""
        self.analysis_dialog.update_types()
        self.analysis_dialog.update_groups()
        self.analysis_dialog.update_cycles()
        self.analysis_dialog.show()

    def show_spindle_help(self):
        self.spindle_help.show()

    def show_slowwave_help(self):
        self.slowwave_help.show()

    def show_evt_analysis_help(self):
        self.evt_analysis_help.show()

    def about(self):
        s = ('<b>WONAMBI Version {version}</b><br />'
             '<p>You can download the latest version at '
             '<a href="https://github.com/wonambi-python/wonambi">'
             'https://github.com/wonambi-python/wonambi</a> '
             'or you can upgrade to the latest release with:'
             '</p><p>'
             '<code>pip install --upgrade wonambi</code>'
             '</p><p>'
             'Copyright &copy; 2013-{year} '
             '<a href="http://www.gpiantoni.com">Gio Piantoni</a>, '
             "Jordan O'Byrne"
             '</p><p>'
             'This program is free software: you can redistribute it '
             'and/or modify it under the terms of the GNU General Public '
             'License as published by the Free Software Foundation, either '
             'version 3 of the License, or (at your option) any later version.'
             '</p><p>'
             'This program is distributed in the hope that it will be useful, '
             'but WITHOUT ANY WARRANTY; without even the implied warranty of '
             'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the '
             'GNU General Public License for more details.'
             '</p><p>'
             'You should have received a copy of the GNU General Public '
             'License along with this program.  If not, see '
             '<a href="http://www.gnu.org/licenses/">'
             'http://www.gnu.org/licenses/</a>.'
             '</p><p>'
             'Other licenses available, contact the author'
             '</p>')
        QMessageBox.about(self, 'WONAMBI', s.format(version=__version__,
                                                    year=now.year))

    def closeEvent(self, event):
        """save the name of the last open dataset."""
        max_dataset_history = self.value('max_dataset_history')
        keep_recent_datasets(max_dataset_history, self.info)

        settings.setValue('window/geometry', self.saveGeometry())
        settings.setValue('window/state', self.saveState())

        event.accept()


app = None


def main():
    global app
    app = QApplication([])

    q = MainWindow()
    q.show()

    app.exec_()



"""Module to select periods of interest, based on number of trials or any of
the axes.

There is some overlap between Select and the Data.__call__(). The main
difference is that Select takes an instance of Data as input and returns
another instance of Data as output, whil Data.__call__() returns the actual
content of the data.

Select should be as flexible as possible. There are quite a few cases, which
will be added as we need them.
"""
from collections import Iterable
from logging import getLogger

from numpy import asarray, empty, linspace, ones, setdiff1d
from scipy.signal import decimate

lg = getLogger(__name__)


def select(data, trial=None, invert=False, **axes_to_select):
    """Define the selection of trials, using ranges or actual values.

    Parameters
    ----------
    data : instance of Data
        data to select from.
    trial : list of int or ndarray (dtype='i'), optional
        index of trials of interest
    **axes_to_select, optional
        Values need to be tuple or list. If the values in one axis are string,
        then you need to specify all the strings that you want. If the values
        are numeric, then you should specify the range (you cannot specify
        single values, nor multiple values). To select only up to one point,
        you can use (None, value_of_interest)
    invert : bool
        take the opposite selection

    Returns
    -------
    instance, same class as input
        data where selection has been applied.
    """
    if trial is not None and not isinstance(trial, Iterable):
        raise TypeError('Trial needs to be iterable.')

    for axis_to_select, values_to_select in axes_to_select.items():
        if (not isinstance(values_to_select, Iterable) or
            isinstance(values_to_select, str)):
            raise TypeError(axis_to_select + ' needs to be iterable.')

    if trial is None:
        trial = range(data.number_of('trial'))
    else:
        trial = trial
        if invert:
            trial = setdiff1d(range(data.number_of('trial')), trial)

    # create empty axis
    output = data._copy(axis=False)
    for one_axis in output.axis:
        output.axis[one_axis] = empty(len(trial), dtype='O')
    output.data = empty(len(trial), dtype='O')

    to_select = {}
    for cnt, i in enumerate(trial):
        lg.debug('Selection on trial {0: 6}'.format(i))
        for one_axis in output.axis:
            values = data.axis[one_axis][i]

            if one_axis in axes_to_select.keys():
                values_to_select = axes_to_select[one_axis]

                if len(values_to_select) == 0:
                    selected_values = ()

                elif isinstance(values_to_select[0], str):
                    selected_values = asarray(values_to_select, dtype='U')

                else:
                    if (values_to_select[0] is None and
                        values_to_select[1] is None):
                        bool_values = ones(len(values), dtype=bool)
                    elif values_to_select[0] is None:
                        bool_values = values < values_to_select[1]
                    elif values_to_select[1] is None:
                        bool_values = values_to_select[0] <= values
                    else:
                        bool_values = ((values_to_select[0] <= values) &
                                       (values < values_to_select[1]))
                    selected_values = values[bool_values]

                if invert:
                    selected_values = setdiff1d(values, selected_values)

                lg.debug('In axis {0}, selecting {1: 6} '
                         'values'.format(one_axis,
                                         len(selected_values)))
                to_select[one_axis] = selected_values
            else:
                lg.debug('In axis ' + one_axis + ', selecting all the '
                         'values')
                selected_values = data.axis[one_axis][i]

            output.axis[one_axis][cnt] = selected_values

        output.data[cnt] = data(trial=i, **to_select)

    return output


def resample(data, s_freq=None, axis='time', ftype='fir', n=None):
    """Downsample the data after applying a filter.

    Parameters
    ----------
    data : instance of Data
        data to downsample
    s_freq : int or float
        desired sampling frequency
    axis : str
        axis you want to apply downsample on (most likely 'time')
    ftype : str
        filter type to apply. The default here is 'fir', like Matlab but unlike
        the default in scipy, because it works better
    n : int
        The order of the filter (1 less than the length for ‘fir’).

    Returns
    -------
    instance of Data
        downsampled data
    """
    output = data._copy()
    ratio = int(data.s_freq / s_freq)

    for i in range(data.number_of('trial')):
        output.data[i] = decimate(data.data[i], ratio,
                                  axis=data.index_of(axis),
                                  zero_phase=True)

        n_samples = output.data[i].shape[data.index_of(axis)]
        output.axis[axis][i] = linspace(data.axis[axis][i][0],
                                        data.axis[axis][i][-1] +
                                        1 / data.s_freq,
                                        n_samples)

    output.s_freq = s_freq

    return output



"""Keep all the files of interest here.

Changing this file will clear the cache of appveyor.
"""
from pathlib import Path

test_path = Path(__file__).resolve().parent

# Data from publicly available datasets
DATA_PATH = test_path / 'data'
axon_abf_file = DATA_PATH / 'axon_abf.abf'
bci2000_file = DATA_PATH / 'bci2000.dat'
psg_file = DATA_PATH / 'PSG.edf'
micromed_file = DATA_PATH / 'micromed.TRC'
generated_file = DATA_PATH / 'edfbrowser_generated_2.edf'

gui_file = psg_file

# Data from private repository
SAMPLE_PATH = DATA_PATH / 'Public'

ANAT_PATH = SAMPLE_PATH / 'anat'
# subsampled version of T1.mgz, to be used as low-res template
template_mri_path = ANAT_PATH / 'T1_subsampled.nii.gz'

ANNOT_PATH = SAMPLE_PATH / 'annot'
chan_path = ANNOT_PATH / 'bert_chan_locs.csv'
annot_domino_path = ANNOT_PATH / 'domino_test_01_EE_stag.txt'
annot_fasst_path = ANNOT_PATH / 'fasst_ACT2007-sh.mat'
annot_alice_path = ANNOT_PATH / 'Staging_Alice.txt'
annot_compumedics_path = ANNOT_PATH / 'Staging_Compumedics.txt'
annot_remlogic_path = ANNOT_PATH / 'Staging_RemLogic.txt'
annot_sandman_path = ANNOT_PATH / 'Staging_Sandman.txt'
annot_psg_path = ANNOT_PATH / 'PSG_scores.xml'

FREESURFER_HOME = SAMPLE_PATH / 'freesurfer'
LUT_path = FREESURFER_HOME / 'FreeSurferColorLUT.txt'
fs_path = FREESURFER_HOME / 'subjects' / 'bert'
surf_path = fs_path / 'surf' / 'lh.pial'

SUBJECTS_DIR = SAMPLE_PATH / 'SUBJECTS_DIR'

IO_PATH = SAMPLE_PATH / 'io'
ktlx_file = IO_PATH / 'xltek'
mff_file = IO_PATH / 'egi.mff'
ns2_file = IO_PATH / 'blackrock' / 'blackrock.ns2'

# Folder where to export data
EXPORTED_PATH = test_path / 'exported'
EXPORTED_PATH.mkdir(exist_ok=True)
annot_file = EXPORTED_PATH / 'annot_scores.xml'
annot_export_file = EXPORTED_PATH / 'annot_scores.csv'
annot_fasst_export_file = EXPORTED_PATH / 'annot_fasst.xml'
exported_chan_path = EXPORTED_PATH / 'grid_chan.sfp'
channel_montage_file = EXPORTED_PATH / 'channel_montage.json'
channel_montage_reref_file = EXPORTED_PATH / 'channel_montage_reref.json'
wonambi_file = EXPORTED_PATH / 'exported.won'

# Store images
DOCS_PATH = test_path.parent / 'docs'
GUI_PATH = DOCS_PATH / 'source' / 'gui' / 'images'
GUI_PATH.mkdir(exist_ok=True)
VIZ_PATH = DOCS_PATH / 'source' / 'viz' / 'images'
VIZ_PATH.mkdir(exist_ok=True)



from os.path import getsize, join
from xml.etree.ElementTree import parse
from datetime import datetime, timedelta, timezone

from numpy import NaN, pad, reshape, zeros

TIMEZONE = timezone.utc
# 24bit precision
DATA_PRECISION = 3

EEG_FILE = 'EEG,Composite,SampleSeries,Composite,MRIAmp,data'
SETTINGS_FILE = 'EEG,Composite,SampleSeries,Composite,MRIAmp,settings'


class Moberg:
    """Basic class to read the data.

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory
    """
    def __init__(self, filename):
        self.filename = filename
        self.n_chan = None
        self.n_smp = None

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header

        Notes
        -----
        the time is probably in "local" Unix Time, which is in the local time
        zone, so we read it as "UTC" (meaning, do not apply timezone
        transformation) and then remove timezone info.
        The only doubt I have is how to interpret the "SystemOffset" time.
        I assume it's in s, and that would fix most of the time zone problems,
        but it does not take into account DST. Or maybe "SystemOffset" is in
        micros and we need to apply the correct time zone to TimeStamp Unix
        time. This needs to be tested with a Moberg system.
        """
        subj_id = str()
        patient = parse(join(self.filename, 'patient.info'))
        for patientname in ['PatientFirstName', 'PatientLastName']:
            subj_id += patient.findall(patientname)[0].text.strip()

        unix_time = int(patient.findall('TimeStamp')[0].text.strip()) / 1e6
        system_offset = int(patient.findall('SystemOffset')[0].text.strip())
        start_time = (datetime.fromtimestamp(unix_time, TIMEZONE) +
                      timedelta(seconds=system_offset)).replace(tzinfo=None)

        s_freq = 256  # could not find it in the text files

        montage = parse(join(self.filename, 'Montage.xml'))
        mont = montage.find('Montage')
        chan_name = [chan.get('lead') for chan in mont.findall('Channel')
                     if chan.get('role') == 'REFERENTIAL_INPUT']

        data_size = getsize(join(self.filename, EEG_FILE))
        n_samples = int(data_size / DATA_PRECISION / len(chan_name))
        self.n_smp = n_samples

        self.n_chan = len(chan_name)
        settings = parse(join(self.filename, SETTINGS_FILE))
        conversion = settings.findall('SampleConversion')[0].text.strip()

        dig_min, dig_max, anl_min, anl_max = [int(x) for x in
                                              conversion.split(',')]

        if dig_max == -dig_min and anl_max == -anl_min:
            self.convertion = lambda dat: dat / dig_max * anl_max
        else:
            self.convertion = lambda dat: ((dat + dig_min) /
                                           (dig_max - dig_min) *
                                           (anl_max - anl_min) + anl_min)

        orig = {'patient': patient,
                'montage': montage,
                'settings': settings,
                }

        return subj_id, start_time, s_freq, chan_name, n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples
        """
        if begsam < 0:
            begpad = -1 * begsam
            begsam = 0
        else:
            begpad = 0

        if endsam > self.n_smp:
            endpad = endsam - self.n_smp
            endsam = self.n_smp
        else:
            endpad = 0

        first_sam = DATA_PRECISION * self.n_chan * begsam
        toread_sam = DATA_PRECISION * self.n_chan * (endsam - begsam)

        with open(join(self.filename, EEG_FILE), 'rb') as f:
            f.seek(first_sam)
            x = f.read(toread_sam)

        dat = _read_dat(x)
        dat = reshape(dat, (self.n_chan, -1), 'F')
        dat = self.convertion(dat[chan, :])
        dat = pad(dat, ((0, 0), (begpad, endpad)),
                  mode='constant', constant_values=NaN)

        return dat

    def return_markers(self):
        """Return all the markers (also called triggers or events).

        Returns
        -------
        list of dict
            where each dict contains 'name' as str, 'start' and 'end' as float
            in seconds from the start of the recordings, and 'chan' as list of
            str with the channels involved (if not of relevance, it's None).

        Raises
        ------
        FileNotFoundError
            when it cannot read the events for some reason (don't use other
            exceptions).
        """
        markers = []
        return markers


def _read_dat(x):
    """read 24bit binary data and convert them to numpy.

    Parameters
    ----------
    x : bytes
        bytes (length should be divisible by 3)

    Returns
    -------
    numpy vector
        vector with the signed 24bit values

    Notes
    -----
    It's pretty slow but it's pretty a PITA to read 24bit as far as I can tell.
    """
    n_smp = int(len(x) / DATA_PRECISION)
    dat = zeros(n_smp)

    for i in range(n_smp):
        i0 = i * DATA_PRECISION
        i1 = i0 + DATA_PRECISION
        dat[i] = int.from_bytes(x[i0:i1], byteorder='little', signed=True)

    return dat
